from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sqlite3, os, json
import numpy as np
import pandas as pd
import datetime as dtm
import multiprocessing

from scipy.optimize import curve_fit
from scipy.special import erfc
from scipy.interpolate import interp1d

from .isotope import Isotope
from .plotter import colors
from .plotter import _init_plot
from .plotter import _close_plot

global DB_CONNECTIONS_DICT
DB_CONNECTIONS_DICT = {}

class Calibration(object):
	"""Calibration is an object for calibrating
	gamma ray data from High-Purity Germanium (HPGe)
	detectors.

	...

	Parameters
	----------
	meta : dict, optional
		Meta attributes for calibration.

	Attributes
	----------
	meta : dict
		Metadata about spectrum. 
	engcal : list
		2 or 3 parameter energy calibration.

	Methods
	-------
	plot(show=True, fit=True, saveas=None, zoom=None, logscale=True, 
			grayscale=False, labels=False, square_fig=False)
		Plots the spectrum. Various options for plotting.
	calibrate(spectra, saveas=None, db=None, auto_calibrate=False)
		Functionality depends on filetype.

	Notes
	-----

	References
	----------

	Examples
	--------

	"""
	def __init__(self, filename=None):
		self.calib = {}
		self._calib_data = {}
		self._default = {'engcal': [0.0, 0.3],
						'effcal': [0.331, 0.158, 0.410, 0.001, 1.476],
						'unc_effcal': [[ 5.038e-02,  3.266e-02, -2.151e-02, -4.869e-05, -7.748e-03],
									 [ 3.266e-02,  2.144e-02, -1.416e-02, -3.416e-05, -4.137e-03],
									 [-2.151e-02, -1.416e-02,  9.367e-03,  2.294e-05,  2.569e-03],
									 [-4.869e-05, -3.416e-05,  2.294e-05,  5.411e-07, -1.165e-04],
									 [-7.748e-03, -4.137e-03,  2.569e-03, -1.165e-04,  3.332e-02]],
						'rescal': [2.0, 4e-4]}
		self.update()

		self._path, self._fnm = None, None
		if filename is not None:
			self._path, self._fnm = os.path.split(filename)
			if self._path in ['',' ']:
				self._path = os.getcwd()
			if self._fnm in ['',' ']:
				raise ValueError('Invalid Filename: {}'.format(filename))

			if os.path.exists(os.path.join(self._path, self._fnm)):
				self.open(os.path.join(self._path, self._fnm))
			else:
				raise ValueError('File does not exist: {}'.format(filename))

	def update(self, meta={}):
		for nm in meta:
			if nm.endswith('cal'):
				self.calib[nm] = meta[nm]
		for nm in ['engcal', 'effcal', 'unc_effcal', 'rescal']:
			if nm not in self.calib:
				self.calib[nm] = self._default[nm]

	@property
	def engcal(self):
		return self.calib['engcal']

	@engcal.setter
	def engcal(self, cal):
		self.calib['engcal'] = cal

	@property
	def effcal(self):
		return self.calib['effcal']

	@effcal.setter
	def effcal(self, cal):
		self.calib['effcal'] = cal

	@property
	def unc_effcal(self):
		return self.calib['unc_effcal']

	@unc_effcal.setter
	def unc_effcal(self, cal):
		self.calib['unc_effcal'] = cal

	@property
	def rescal(self):
		return self.calib['rescal']

	@rescal.setter
	def rescal(self, cal):
		self.calib['rescal'] = cal
	
	def eng(self, idx, *cal):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		idx = np.asarray(idx)
		cal = cal if len(cal) else self.calib['engcal']
		if len(cal)==2:
			return cal[0]+cal[1]*idx
		elif len(cal)==3:
			return cal[0]+cal[1]*idx+cal[2]*idx**2

	def eff(self, energy, *c):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		energy = np.asarray(energy)
		c = c if len(c) else self.calib['effcal']
		if len(c)==3:
			return c[0]*np.exp(-c[1]*energy**c[2])
		elif len(c)==5:
			return c[0]*np.exp(-c[1]*energy**c[2])*(1.0-np.exp(-c[3]*energy**c[4]))
			
	def unc_eff(self, energy, c=None, u=None):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		energy = np.asarray(energy)
		if c is None or u is None:
			c, u = self.calib['effcal'], self.calib['unc_effcal']
		eps = 1E-8
		var= np.zeros(len(energy)) if energy.shape else 0.0
		for n in range(len(c)):
			for m in range(n, len(c)):
				if not np.isfinite(u[n][m]):
					return np.inf
				c_n, c_m = list(c), list(c)
				c_n[n], c_m[m] = c_n[n]+eps, c_m[m]+eps
				par_n = (self.eff(energy, *c_n)-self.eff(energy, *c))/eps
				par_m = (self.eff(energy, *c_m)-self.eff(energy, *c))/eps
				var += u[n][m]*par_n*par_m*(2.0 if n!=m else 1.0)
		return np.sqrt(var)

	def res(self, idx, *cal):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		idx = np.asarray(idx)
		cal = cal if len(cal) else self.calib['rescal']
		if len(cal)==1:
			return cal[0]*np.sqrt(idx)
		elif len(cal)==2:
			return cal[0]+cal[1]*idx

	def map_idx(self, energy, *cal):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		energy = np.asarray(energy)
		cal = cal if len(cal) else self.calib['engcal']
		if len(cal)==3:
			if cal[2]!=0.0:
				return np.array(np.rint(0.5*(np.sqrt(cal[1]**2-4.0*cal[2]*(cal[0]-energy))-cal[1])/cal[2]), dtype=np.int32)
		return np.array(np.rint((energy-cal[0])/cal[1]), dtype=np.int32)

	def calibrate(self, spectra, saveas=None, db=None, auto_calibrate=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		shelves = []
		sps = []
		
		for sp in spectra:
			if type(sp)==str:
				sp = Spectrum(sp, db)
			sps.append(sp)
			if ('A0' in sp.meta and 'ref_date' in sp.meta):
				if 'shelf' not in sp.meta:
					sp.meta = {'shelf':None}
				shelves.append(sp.meta['shelf'])
		cal_spec = [sp for sp in spectra if ('A0' in sp.meta and 'ref_date' in sp.meta)]
		if len(cal_spec)==0:
			raise ValueError('Cannot calibrate: A0 and ref_date required for calibration.')

		if auto_calibrate:
			for sp in cal_spec:
				sp.auto_calibrate(**{'_cb_cal':False})
		p0_engcal = np.average([s.cb.engcal for s in cal_spec if len(s.cb.engcal)==len(cal_spec[0].cb.engcal)], axis=0)
		engcal = self._calibrate_energy(cal_spec, p0_engcal)
		rescal = self._calibrate_resolution(cal_spec, np.average([s.cb.rescal for s in cal_spec], axis=0))
		for sp in sps:
			sp.meta = {'engcal':engcal, 'rescal':rescal}
		self.update({'engcal':engcal, 'rescal':rescal})

		shelves = {sh:[s for s in cal_spec if s.meta['shelf']==sh] for sh in set(shelves)}
		for shelf in shelves:
			if 'effcal' not in self._calib_data:
				self._calib_data['effcal'] = []

			shelf_specs = shelves[shelf]
			p0_effcal = np.average([s.cb.effcal for s in shelf_specs if len(s.cb.effcal)==len(shelf_specs[0].cb.effcal)], axis=0)
			effcal, unc_effcal = self._calibrate_efficiency(shelf_specs, p0_effcal)
			for sp in sps:
				if shelf==sp.meta['shelf']:
					sp.meta = {'effcal':effcal, 'unc_effcal':unc_effcal}
					self.update({'effcal':effcal, 'unc_effcal':unc_effcal})
					if saveas is not None:
						if type(saveas)==str:
							saveas = [saveas]
						sp.save(*saveas)

	def _calibrate_energy(self, spectra, p0):
		E, chn, unc_chn = [], [], []
		for sp in spectra:
			sp.meta = {'engcal':p0}
			for f in sp.fits:
				E += f['gm'][:,0].tolist()
				chn += [f['fit'][i] for i in f['ix']['mu']]
				unc_chn += [np.sqrt(f['unc'][i][i]) for i in f['ix']['mu']]

		E, chn, unc_chn = np.array(E), np.array(chn), np.array(unc_chn)
		idx = np.where((chn>unc_chn)&(np.isfinite(unc_chn))&(unc_chn>1E-3))
		E, chn, unc_chn = E[idx], chn[idx], unc_chn[idx]

		p0 = p0 if len(p0)==3 else [p0[0], p0[1], 0.0]
		fit, unc = curve_fit(self.eng, chn, E, sigma=self.eng(unc_chn), p0=p0)

		self._calib_data['engcal'] = {'fit':fit, 'unc':unc, 'x':E, 'y':chn, 'yerr':unc_chn}
		return fit

	def _calibrate_efficiency(self, spectra, p0):
		E, eff, unc_eff = [], [], []
		lms = {}
		for sp in spectra:
			sp.meta = {'effcal':p0}
			p = sp.peaks
			E += p['energy'].tolist()
			if type(sp.meta['A0'])==list:
				A0 = np.array([sp.meta['A0'][sp.meta['istp'].index(i)] for i in p['isotope']])
			else:
				A0 = np.array([sp.meta['A0'] for i in p['isotope']])
			if type(sp.meta['ref_date'])==list:
				rd = [sp.meta['ref_date'][sp.meta['istp'].index(i)] for i in p['isotope']]
				d0 = [dtm.datetime.strptime(r, '%m/%d/%Y %H:%M:%S') for r in rd]
				t_d = np.array([(sp.meta['start_time']-d).total_seconds() for d in d0])
			else:
				rd = sp.meta['ref_date']
				d0 = dtm.datetime.strptime(rd, '%m/%d/%Y %H:%M:%S')
				t_d = np.array([(sp.meta['start_time']-d0).total_seconds() for i in p['isotope']])

			lm = []
			for i in p['isotope']:
				if i not in lms:
					lms[i] = Isotope(i).decay_const(unc=True)
				lm.append(lms[i])
			lm = np.array(lm)

			ef = p['counts']*lm[:,0]/((1.0-np.exp(-lm[:,0]*sp.meta['real_time']))*np.exp(-lm[:,0]*t_d)*p['intensity']*A0*(sp.meta['live_time']/sp.meta['real_time']))
			unc_eff += (ef*np.sqrt((p['unc_counts']/p['counts'])**2+(p['unc_intensity']/p['intensity'])**2+(lm[:,1]/lm[:,0])**2+0.01**2)).tolist()
			eff += ef.tolist()

		E, eff, unc_eff = np.array(E), np.array(eff), np.array(unc_eff)
		idx = np.where((eff>unc_eff)&(unc_eff>0.0)&(np.isfinite(unc_eff)))
		E, eff, unc_eff = E[idx], eff[idx], unc_eff[idx]
		p0 = p0.tolist() if len(p0)==5 else p0.tolist()+[0.001, 1.476]
		p0[0] = max([min([p0[0]*np.average(np.array(eff)/self.eff(np.array(E), *p0), weights=(self.eff(np.array(E), *p0)/np.array(unc_eff))**2),99.9]),0.001])

		bounds = ([0.0, 0.0, -1.0, 0.0, -2.0], [100.0, 3.0, 3.0, 0.5, 3.0])
		if any([sp.fit_config['xrays'] for sp in spectra]):
			try:
				fit3, unc3 = curve_fit(self.eff, E, eff, sigma=unc_eff, p0=p0[:3], bounds=(bounds[0][:3], bounds[1][:3]))
				fit5, unc5 = curve_fit(self.eff, E, eff, sigma=unc_eff, p0=fit3.tolist()+p0[3:], bounds=bounds)
				
				chi5 = np.sum((eff-self.eff(E, *fit5))**2/unc_eff**2)
				chi3 = np.sum((eff-self.eff(E, *fit3))**2/unc_eff**2)
				## Invert to find which is closer to one
				chi5 = chi5 if chi5>1.0 else 1.0/chi5
				chi3 = chi3 if chi3>1.0 else 1.0/chi3
				fit, unc = (fit3, unc3) if chi3<=chi5 else (fit5, unc5)
			except:
				fit, unc = curve_fit(self.eff, E, eff, sigma=unc_eff, p0=p0[:3], bounds=(bounds[0][:3], bounds[1][:3]))

		else:
			fit, unc = curve_fit(self.eff, E, eff, sigma=unc_eff, p0=p0[:3], bounds=(bounds[0][:3], bounds[1][:3]))

		self._calib_data['effcal'].append({'fit':fit, 'unc':unc, 'x':E, 'y':eff, 'yerr':unc_eff, 'shelf':spectra[0].meta['shelf']})
		return fit, unc

	def _calibrate_resolution(self, spectra, p0):
		mu, sig, unc_sig = [], [], []
		for sp in spectra:
			sp.meta = {'rescal':p0}
			for f in sp.fits:
				mu += [f['fit'][i] for i in f['ix']['mu']]
				sig += [f['fit'][i] for i in f['ix']['sig']]
				u_sig = [f['unc'][i][i] for i in f['ix']['sig']]
				u_mu = [f['unc'][i][i] for i in f['ix']['mu']]
				unc_sig += [np.sqrt(s+0.25*p0[0]*u_mu[n]/mu[-1*(len(u_mu)-n)]) for n,s in enumerate(u_sig)]
		mu, sig, unc_sig = np.array(mu), np.array(sig), np.array(unc_sig)
		idx = np.where((sig>unc_sig)&(np.isfinite(unc_sig))&(unc_sig>1E-3))
		mu, sig, unc_sig = mu[idx], sig[idx], unc_sig[idx]
		fit, unc = curve_fit(self.res, mu, sig, sigma=unc_sig, p0=p0, bounds=([-np.inf, 0.0],[np.inf, np.inf]))

		self._calib_data['rescal'] = {'fit':fit, 'unc':unc, 'x':mu, 'y':sig, 'yerr':unc_sig}
		return fit

	def open(self, filename):
		if filename.endswith('.json'):
			with open(filename) as f:
				js = json.load(f)

			for c in ['engcal','rescal','effcal','unc_effcal']:
				if c in js:
					self.update({c:np.array(js[c])})
			if '_calib_data' in js:
				for c in ['engcal','rescal']:
					if c in js['_calib_data']:
						self._calib_data[c] = {str(i):np.array(js['_calib_data'][c][i]) for i in js['_calib_data'][c]}
				if 'effcal' in js['_calib_data']:
					self._calib_data['effcal'] = []
					for e in js['_calib_data']['effcal']:
						self._calib_data['effcal'].append({str(i):(np.array(e[i]) if i!='shelf' else (str(e[i]) if e[i] is not None else None)) for i in e})


	def saveas(self, filename):
		if filename.endswith('.json'):
			js = {}
			for c in ['engcal','rescal','effcal','unc_effcal']:
				if type(self.calib[c])==list:
					js[c] = self.calib[c]
				else:
					js[c] = self.calib[c].tolist()
			js['_calib_data'] = {}
			for cl in ['engcal','rescal']:
				if cl in self._calib_data:
					js['_calib_data'][cl] = {i:self._calib_data[cl][i].tolist() for i in self._calib_data[cl]}
			if 'effcal' in self._calib_data:
				js['_calib_data']['effcal'] = []
				for e in self._calib_data['effcal']:
					js['_calib_data']['effcal'].append({i:(e[i].tolist() if i!='shelf' else e[i]) for i in e})

			with open(filename, 'w') as f:
				json.dump(js, f, indent=4)

	def plot(self, **kwargs):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		cm = colors(aslist=True)
		cm_light = colors(shade='light', aslist=True)
		f, ax = _init_plot(N_plots=3, figsize=(12.8, 4.8))

		if 'engcal' in self._calib_data:
			d = self._calib_data['engcal']
			ax[0].errorbar(d['x'], d['y'], yerr=d['yerr'], ls='None', marker='o')
			x = np.arange(min(d['x']), max(d['x']), 0.1)
			ax[0].plot(x, self.map_idx(x, *d['fit']))
			ax[0].set_xlabel('Energy (keV)')
			ax[0].set_ylabel('ADC Channel')
			ax[0].set_title('Energy Calibration')

		if 'rescal' in self._calib_data:
			d = self._calib_data['rescal']
			ax[1].errorbar(d['x'], d['y'], yerr=d['yerr'], ls='None', marker='o')
			x = np.arange(min(d['x']), max(d['x']), 0.1)
			ax[1].plot(x, self.res(x, *d['fit']))
			ax[1].set_xlabel('ADC Channel')
			ax[1].set_ylabel('Peak Width')
			ax[1].set_title('Resolution Calibration')
		
		if 'effcal' in self._calib_data:
			for N,d in enumerate(sorted(self._calib_data['effcal'], key=lambda h: h['shelf'])):
				ax[2].errorbar(d['x'], d['y'], yerr=d['yerr'], ls='None', marker='o', color=cm[N%len(cm)])
				x = np.arange(min(d['x']), max(d['x']), 0.1)
				ax[2].plot(x, self.eff(x, *d['fit']), color=cm[N%len(cm)], label=d['shelf'])
				low = self.eff(x, *d['fit'])-self.unc_eff(x, d['fit'], d['unc'])
				high = self.eff(x, *d['fit'])+self.unc_eff(x, d['fit'], d['unc'])
				ax[2].fill_between(x, low, high, facecolor=cm_light[N%len(cm)], alpha=0.5)
			ax[2].set_xlabel('Energy (keV)')
			ax[2].set_ylabel('Efficiency')
			ax[2].set_title('Efficiency Calibration')
			if any([d['shelf'] for d in self._calib_data['effcal']]):
				ax[2].legend(loc=0)

		return _close_plot(f, None, default_log=False, **kwargs)





def _parallel_fit(dat):
	x, y, fit_config, snip, p0, bounds = dat['x'], dat['y'], dat['fit_config'], dat['snip'], dat['p0'], dat['bounds']

	def _multiplet(x, *args):

		def _peak(x, A, mu, sig, R, alpha, step):
			r2 = 1.41421356237
			return A*np.exp(-0.5*((x-mu)/sig)**2)+R*A*np.exp((x-mu)/(alpha*sig))*erfc((x-mu)/(r2*sig)+1.0/(r2*alpha))+step*A*erfc((x-mu)/(r2*sig))

		if fit_config['bg_fit']:
			if fit_config['quad_bg']:
				b, peak = 3, args[0]+args[1]*x+args[2]*x**2
			else:
				b, peak = 2, args[0]+args[1]*x
		else:
			b, peak = 0, snip.copy()
		R, alpha, step = fit_config['R'], fit_config['alpha'], fit_config['step']
		if fit_config['skew_fit']:
			if fit_config['step_fit']:
				for n in range(int((len(args)-b)/6)):
					peak += _peak(x,*args[6*n+b:6*n+6+b])
			else:
				for n in range(int((len(args)-b)/5)):
					peak += _peak(x,*(args[5*n+b:5*n+5+b]+(step,)))
		else:
			if fit_config['step_fit']:
				for n in range(int((len(args)-b)/4)):
					peak += _peak(x,*(args[4*n+b:4*n+b+3]+(R, alpha)+(args[4*n+3+b])))
			else:
				for n in range(int((len(args)-b)/3)):
					peak += _peak(x,*(args[3*n+b:3*n+b+3]+(R,alpha,step)))

		return peak

	try:
		fit, unc = curve_fit(_multiplet, x, y, p0=p0, bounds=bounds, sigma=np.sqrt(y+0.1))
		converged = True
	except:
		fit, unc = np.array(p0), np.inf*np.ones((len(p0), len(p0)))
		converged = False
	dat['fit'], dat['unc'], dat['converged'] = fit, unc, converged
	return dat


class Spectrum(object):
	"""Spectrum is a class for fitting and plotting
	gamma ray data from High-Purity Germanium (HPGe)
	detectors.

	...


	Parameters
	----------
	filename : str
		Path to .Spe file.
	db : str, optional
		Path to sqlite database

	Attributes
	----------
	fits : list
		List of peaks (PeakFit class) found in spectrum fit.
	meta : dict
		Metadata about spectrum. 
	hist : np.ndarray
		1D histogram of pulse amplitudes (energy).

	Methods
	-------
	plot(show=True, fit=True, saveas=None, zoom=None, logscale=True, 
			grayscale=False, labels=False, square_fig=False)
		Plots the spectrum. Various options for plotting.
	save(*saveas)
		Functionality depends on filetype.
	summarize(printout=True, saveas=None)
		Prints and/or saves summary of peaks/isotopes.

	Notes
	-----

	References
	----------

	Examples
	--------

	"""

	def __init__(self, filename=None, db=None):
		self._default_params()

		if filename is not None:
			self._path, self._fnm = os.path.split(filename)
			if self._path in ['',' ']:
				self._path = os.getcwd()
			if self._fnm in ['',' ']:
				raise ValueError('Invalid Spectrum Filename: {}'.format(filename))

			if os.path.exists(os.path.join(self._path, self._fnm)):
				if self._fnm.endswith('.Spe'):
					self._from_Spe()
				elif self._fnm.endswith('.Chn'):
					self._from_Chn()
			else:
				raise ValueError('File does not exist: {}'.format(filename))

		self._check_db(db)
		self.cb.update(self._meta)


	def _default_params(self):
		self._path, self._fnm = None, None
		self.db, self.db_connection, self.db_fnm = None, None, None
		self._hist = np.zeros(2, dtype=np.int32)
		self._meta = {'fit_config':{'snip_adj':1.0, 'R':0.1, 'alpha':0.9, 
									'step':0.0, 'bg_fit':False, 'skew_fit':False, 
									'step_fit':False, 'quad_bg':False, 'SNR_cut':4.0, 
									'A_bound':10.0, 'mu_bound':1.5, 'sig_bound':1.5,
									'xrays':False, 'pk_width':7.5, 'E_min':75.0, 
									'I_min':0.05,'threads':multiprocessing.cpu_count()},
						'istp':[], 'shelf':None, 'spec_id':1}
		self._cb = Calibration()
		for nm in self.cb.calib:
			self._meta[nm] = self.cb.calib[nm]
		self._gamma_list = None
		self._fits, self._peaks = None, None

	def _get_db_params(self):
		if self.db is not None:
			from ast import literal_eval
			tables = [str(i[0]) for i in self.db.execute('SELECT name FROM sqlite_master WHERE type="table"')]

			if 'spectra' not in tables:
				q = 'CREATE TABLE `spectra` (`spec_id` INTEGER, `filename` TEXT, `file_path` TEXT, '
				q += '`start_time` TEXT, `live_time` REAL, `real_time` REAL, `meta` TEXT, `fit_config` TEXT, '
				q += '`engcal` TEXT, `effcal` TEXT, `unc_effcal` TEXT, `rescal` TEXT);'
				self.db.execute(q)
				self.db_connection.commit()

			q = list(self.db.execute('SELECT * FROM spectra WHERE filename LIKE ? AND file_path LIKE ?', ('%'+self._fnm+'%', '%'+self._path+'%')))
			if len(q):
				self.meta = {'engcal':literal_eval(str(q[0][8])), 'effcal':literal_eval(str(q[0][9])), 
							'unc_effcal':np.array(literal_eval(str(q[0][10]).replace('inf',"'inf'")), dtype=np.float64), 
							'rescal':literal_eval(str(q[0][11]))}
				self._meta['spec_id'] = q[0][0]
			else:
				ids = [i[0] for i in self.db.execute('SELECT spec_id FROM spectra')]
				self._meta['spec_id'] = max(ids)+1 if len(ids) else 1
				
				vals = [self.meta['spec_id'], self._fnm, self._path, dtm.datetime.strftime(self.meta['start_time'], '%m/%d/%Y %H:%M:%S'),
						self.meta['live_time'], self.meta['real_time'],
						str({i:self.meta[i] for i in self.meta if i not in ['start_time', 'fit_config','engcal','effcal','unc_effcal','rescal']}),
						str({i:self.fit_config[i] for i in self.fit_config if i not in []}),
						str(list(self.meta['engcal'])), str(list(self.meta['effcal'])), 
						(str(self.meta['unc_effcal']) if type(self.meta['unc_effcal'])==list else str(self.meta['unc_effcal'].tolist())),
						str(list(self.meta['rescal']))]
				self.db.execute('INSERT INTO spectra VALUES('+','.join(12*['?'])+')', tuple(vals))
				self.db_connection.commit()

	def _check_db(self, db=None):
		if db is not None:
			path, fnm = os.path.split(db)
			if path in ['',' ']:
				path = os.getcwd()
			if fnm in ['',' ']:
				raise ValueError('Invalid db Filename: {}'.format(db))
			self.db_fnm = os.path.join(path, fnm)

			global DB_CONNECTIONS_DICT
			if os.path.exists(self.db_fnm):
				if self.db_fnm not in DB_CONNECTIONS_DICT:
					DB_CONNECTIONS_DICT[self.db_fnm] = sqlite3.connect(self.db_fnm)
				self.db_connection = DB_CONNECTIONS_DICT[self.db_fnm]
				self.db = self.db_connection.cursor()
			else:
				print('WARNING: DB {} does not exist, creating new file.'.format(fnm))
				from sqlite3 import Error
				try:
					self.db_connection = sqlite3.connect(self.db_fnm)
					DB_CONNECTIONS_DICT[self.db_fnm] = self.db_connection
					self.db = self.db_connection.cursor()
				except Error as e:
					print(e)
			self._get_db_params()

	def _update_db(self, meta):
		if self.db is not None:
			if meta:
				meta_str = str({i:self.meta[i] for i in self.meta if i not in ['start_time','fit_config','engcal','effcal','unc_effcal','rescal']})
				self.db.execute('UPDATE spectra SET meta=? WHERE spec_id=?', (meta_str, self.meta['spec_id']))
				fit_config_str = str({i:self.fit_config[i] for i in self.fit_config if i not in []})
				self.db.execute('UPDATE spectra SET fit_config=? WHERE spec_id=?', (fit_config_str, self.meta['spec_id']))
				vals = [str(list(self.meta['engcal'])), str(list(self.meta['effcal'])), 
						(str(self.meta['unc_effcal']) if type(self.meta['unc_effcal'])==list else str(self.meta['unc_effcal'].tolist())),
						str(list(self.meta['rescal'])), self.meta['spec_id']]
				self.db.execute('UPDATE spectra SET engcal=?, effcal=?, unc_effcal=?, rescal=? WHERE spec_id=?', tuple(vals))
			elif self._fits is not None:
				if 'peaks' in [str(i[0]) for i in self.db.execute('SELECT name FROM sqlite_master WHERE type="table"')]:
					self.db.execute('DELETE FROM peaks WHERE spec_id=?',(self.meta['spec_id'],))
				self.peaks.to_sql('peaks', self.db_connection, if_exists='append', index=False)
			self.db_connection.commit()

	@property
	def cb(self):
		return self._cb

	@cb.setter
	def cb(self, _cb):
		import copy
		self._cb = copy.deepcopy(_cb)
		if 'effcal' in _cb._calib_data:
			shelf = self.meta['shelf'] if 'shelf' in self.meta else None
			for sh in _cb._calib_data['effcal']:
				if sh['shelf']==shelf:
					self.meta = {'effcal':sh['fit'], 'unc_effcal':sh['unc']}	

	@property
	def hist(self):
		return self._hist

	@hist.setter
	def hist(self, hist_array):
		self._hist = np.asarray(hist_array, dtype=np.int32)
		self.channels = np.arange(len(self._hist))
		self._snip = self.snip_bg()
		self.snip = interp1d(self.channels, self._snip)


	@property
	def meta(self):
		return self._meta

	@meta.setter
	def meta(self, meta_dict):
		for nm in meta_dict:
			if nm.endswith('cal'):
				self.cb.calib[nm] = meta_dict[nm]
			self._meta[nm] = meta_dict[nm]
			if nm=='istp':
				self._fits, self._peaks, self._gamma_list = None, None, None
		self._update_db(True)
	
	@property
	def fit_config(self):
		return self._meta['fit_config']
	
	@fit_config.setter
	def fit_config(self, config_dict):
		self._fits, self._peaks, self._gamma_list = None, None, None
		for nm in config_dict:
			self._meta['fit_config'][nm] = config_dict[nm]

	def __add__(self, other):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""
		
		if not type(other)==Spectrum:
			raise ValueError('Cannot add spectrum to type {}'.format(type(other)))

		if self.meta['start_time']==other.meta['start_time']:
			alpha = np.sum(self.hist)/float(np.sum(other.hist))
			dead_time = alpha*(self.meta['real_time']-self.meta['live_time'])+(1.0-alpha)*(other.meta['real_time']-other.meta['live_time'])
			self.meta['real_time'] = alpha*self.meta['real_time']+(1.0-alpha)*other.meta['real_time']
			self.meta['live_time'] = self.meta['real_time']-dead_time
		else:
			self.meta['live_time'] += other.meta['live_time']
			self.meta['real_time'] += other.meta['real_time']

		if len(self.hist)==len(other.hist):
			if len(self.cb.engcal)==len(other.cb.engcal):
				if len(np.where(self.cb.engcal==other.cb.engcal)[0])==len(self.cb.engcal):
					self.hist += other.hist
					return self

		other_bins = other.cb.eng(np.arange(-0.5, len(other.hist)+0.5, 1.0))
		dNdE = np.array(other.hist, dtype=np.float64)/(other_bins[1:]-other_bins[:-1])
		f = interp1d(other.cb.eng(np.arange(len(other.hist))), dNdE, bounds_error=False, fill_value=0.0)

		bins = self.cb.eng(np.arange(-0.5, len(self.hist)+0.5, 1.0))
		edges = np.append(bins, bins).reshape((2,len(bins))).T.flatten()[1:-1].reshape((len(bins)-1, 2)).T
		if np.__version__>='1.16.0':
			e_grid = np.linspace(edges[0], edges[1], num=10).T
		else:
			e_grid = np.array([np.linspace(edges[0][n], edges[1][n], num=10) for n in range(len(edges[0]))])
		N = np.asarray(np.trapz(f(e_grid), e_grid, axis=1), dtype=np.int64)
		self.hist += np.random.poisson(N)

		return self

	def rebin(self, N_bins):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		L = len(self.hist)
		if N_bins>L:
			raise ValueError('N_bins: {0} must not be greater than current value: {1}'.format(N_bins, L))
		r = int(round(L/float(N_bins)))
		self.hist = np.sum(self.hist.reshape((int(L/r), r)), axis=1)
		ec = self.cb.engcal
		self.meta = {'engcal':[ec[0], ec[1]*r]+([ec[2]*r**2] if len(ec)==3 else [])}
		rc = self.cb.rescal
		self.meta = {'rescal':([rc[0]*np.sqrt(r)] if len(rc)==1 else [rc[0], rc[1]*r])}

	def exp_smooth(self, x, alpha=0.3):
		N = int(2.0/alpha)-1
		wts = np.array([alpha*(1.0-alpha)**abs(N-i) for i in range(2*N+1)])
		y = np.concatenate((np.ones(N)*x[0], x, np.ones(N)*x[-1]))
		return np.dot(wts/np.sum(wts), np.take(y, [np.arange(i,len(x)+i) for i in range(2*N+1)]))

	def snip_bg(self):
		adj = self.fit_config['snip_adj']

		x, dead = self.channels, int(7.5*adj*self.cb.res(len(self._hist)))
		V_i, L = np.log(np.log(np.sqrt(self._hist+1.0)+1.0)+1.0), len(x)-dead
		while self._hist[dead]==0:
			dead += 1
		
		for M in np.linspace(0, 7.5*adj, 10):
			l, h = np.array(x-M*self.cb.res(x), dtype=np.int32), np.array(x+M*self.cb.res(x), dtype=np.int32)
			V_i[dead:L] = np.minimum(V_i[dead:L],0.5*(V_i[l[dead:L]]+V_i[h[dead:L]]))

		snip = (np.exp(np.exp(V_i)-1.0)-1.0)**2-1.0
		snip += adj*1.5*np.sqrt(snip+0.01)

		R1, R2 = 0.3/(adj*self.cb.res(0.1*L)), 0.3/(adj*self.cb.res(0.9*L))
		wt1, wt2 = np.flip(x, 0)/x[-1], x/x[-1]
		return wt1*self.exp_smooth(snip, R1) + wt2*self.exp_smooth(snip, R2)

	@property
	def gamma_list(self):
		if self._gamma_list is None:
			if 'istp' in self._meta:
				xrays, E_min, I_min = self.fit_config['xrays'], self.fit_config['E_min'], self.fit_config['I_min']
				gammas = [Isotope(i).gammas(E_lim=[E_min, self.cb.eng(0.99*len(self._hist))], I_lim=[I_min, None], xrays=xrays) for i in self._meta['istp']]
				self._gamma_list = [(np.array([g['E'], g['I'], g['dI']]).T)*np.array([1.0, 0.01, 0.01]) for g in gammas]
			else:
				self._gamma_list = []
		return self._gamma_list

	def _guess_forward_activity(self, *engcal):
		x, Y, p0_A = self.channels, [], []
		R, alpha, step = self.fit_config['R'], self.fit_config['alpha'], self.fit_config['step']
		clip = np.where(self._hist>self._snip, self._hist-self._snip, 0.0)
		W = self.fit_config['pk_width']
		for gm in self.gamma_list:
			L = len(self._hist)
			y = np.zeros(L)
			if len(gm):
				idx = self.cb.map_idx(gm[:,0], *engcal)
				sig = self.cb.res(idx)
				norm = self.cb.eff(gm[:,0])*gm[:,1]/sig
				idx, sig, norm = idx[idx<L], sig[idx<L], norm[idx<L]
				p0_A.append(np.exp(np.average(np.log((clip[idx]/norm)+1.0), weights=np.sqrt(norm)))-1.0)
				if p0_A[-1]<0:
					p0_A[-1] = 0.0
				l, h = np.array(idx-W*sig, dtype=np.int32), np.array(idx+W*sig, dtype=np.int32)
				for m,i in enumerate(idx):
					if l[m]>0 and h[m]<len(x):
						y[l[m]:h[m]] += self._peak(x[l[m]:h[m]], norm[m], i, sig[m], R, alpha, step)
				Y.append(y)
		return p0_A, np.array(Y)

	def _fit_forward_activity(self):
		p0_A, Y = self._guess_forward_activity()
		try:
			return curve_fit(lambda x, *A: self._snip+np.dot(A, Y), self.channels, self._hist, p0=p0_A)
		except:
			return p0_A, np.ones((len(p0_A), len(p0_A)))*np.inf

	def auto_calibrate(self, guess=[], data=[], **param):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		from scipy.optimize import differential_evolution
		guess = guess if len(guess) else list(self.cb.engcal)
		if len(data):
			data = np.asarray(data, dtype=np.float64)
			if len(data)==1:
				guess = [0.0, data[0][1]/data[0][0]]
			else:
				N = 2 if len(data)<5 else 3
				M = np.column_stack([data[:,0]**m for m in range(N)])
				b = np.array([np.sum(data[:,0]**m*data[:,1]) for m in range(N)])
				M_inv = np.linalg.inv(np.dot(M.T, M))
				guess = np.dot(M_inv, b).tolist()
		guess = guess if len(guess)==3 else list(guess)+[0.0]
		obj = lambda m: np.average(np.sqrt(np.absolute(self._hist-self._snip-np.dot(*self._guess_forward_activity(*[guess[0], m[0], guess[2]])))))
		self.meta = {'engcal': [guess[0], differential_evolution(obj, [(0.995*guess[1], 1.005*guess[1])]).x[0], guess[2]]}

		if ('A0' in self.meta and 'ref_date' in self.meta) and (param['_cb_cal'] if '_cb_cal' in param else True):
			eff, u_eff, res = self.cb.effcal, self.cb.unc_effcal, self.cb.rescal
			try:
				self.cb.calibrate([self])
			except Exception as e:
				print('Error:', e)
				self.meta = {'engcal':guess, 'effcal':eff, 'unc_effcal':u_eff, 'rescal':res}
				try:
					self.cb.calibrate([self])
				except Exception as e:
					print('Error:', e)
					self.meta = {'engcal':guess, 'effcal':eff, 'unc_effcal':u_eff, 'rescal':res}
					print('WARNING: auto calibrate failed. Setting engcal to guess.')

		elif len(self.meta['istp']) and (param['_cb_cal'] if '_cb_cal' in param else True):
			eng, res = self.cb.engcal, self.cb.rescal
			fit_config = self.fit_config.copy()
			self.fit_config = {'bg_fit':False, 'skew_fit':False}
			calib_data = self.cb._calib_data.copy()
			try:
				self.cb._calib_data['engcal'] = {}
				self.cb._calib_data['rescal'] = {}
				self.meta = {'engcal':self.cb._calibrate_energy([self],eng), 'rescal':self.cb._calibrate_resolution([self],res)}
			except Exception as e:
				print('Error:', e)
				self.meta = {'engcal':guess, 'rescal':res}
				self.cb._calib_data = calib_data
				print('WARNING: auto calibrate failed. Setting engcal to guess.')
			self.fit_config = fit_config
			self._fits, self._peaks = None, None


	def _chi2(self, fit, lh):
		non_zero = np.where(self.hist[lh[0]:lh[1]]>0)
		dof = float(len(non_zero[0])-len(fit)-1)
		if dof==0:
			return np.inf
		resid = self.hist[lh[0]:lh[1]][non_zero]-self.multiplet(self.channels[lh[0]:lh[1]][non_zero], *fit)
		return np.sum(resid**2/self.hist[lh[0]:lh[1]][non_zero])/dof
		
	def _unc_calc(self, fn, p0, cov):
		var, eps = 0.0, 1E-8
		for n in range(len(p0)):
			for m in range(n, len(p0)):
				c_n, c_m = list(p0), list(p0)
				c_n[n], c_m[m] = c_n[n]+eps, c_m[m]+eps
				par_n = (fn(*c_n)-fn(*p0))/eps
				par_m = (fn(*c_m)-fn(*p0))/eps
				var += cov[n][m]*par_n*par_m*(2.0 if n!=m else 1.0)
		return var
	
	def _counts(self, fit, cov, lh):
		cfg = self.fit_config
		L = 3+2*int(cfg['skew_fit'])+int(cfg['step_fit'])
		M = int(cfg['bg_fit'])*(2+int(cfg['quad_bg']))
		N_cts, unc_N = [], []

		skew_fn = lambda A, R, alpha, sig: 2*A*R*alpha*sig*np.exp(-0.5/alpha**2)
		min_skew_fn = lambda A, sig: 2*A*cfg['R']*cfg['alpha']*sig*np.exp(-0.5/cfg['alpha']**2)
		pk_fn = lambda A, sig: 2.506628*A*sig

		for m in range(int((len(fit)-M)/L)):
			i = L*m+M
			p_i = [i, i+2]
			s_i = [i, i+3, i+4, i+2]
			if cfg['skew_fit']:
				N_cts.append(pk_fn(*fit[p_i])+skew_fn(*fit[s_i]))
			else:
				N_cts.append(pk_fn(*fit[p_i])+min_skew_fn(*fit[p_i]))
			if cov is not None:
				if not np.isinf(cov[i][i]):
					if cfg['skew_fit']:
						skew_unc = self._unc_calc(skew_fn, fit[s_i], cov[np.ix_(s_i, s_i)])
					else:
						skew_unc = self._unc_calc(min_skew_fn, fit[p_i], cov[np.ix_(p_i, p_i)])
					unc_N.append(np.sqrt(self._unc_calc(pk_fn, fit[p_i], cov[np.ix_(p_i, p_i)])+skew_unc))
				else:
					unc_N.append(np.inf)
			else:
				unc_N.append(np.inf)
		return np.array(N_cts), np.array(unc_N)
	
	def _decays(self, N, unc_N, gm):
		if 'live_time' not in self.meta or 'real_time' not in self.meta:
			raise ValueError('Need real-time/live-time to compute decay-rate.')

		D = N/(gm[:,1]*self.cb.eff(gm[:,0])*(self.meta['live_time']/self.meta['real_time']))
		unc_D = np.sqrt((unc_N/N)**2+(self.cb.unc_eff(gm[:,0])/self.cb.eff(gm[:,0]))**2+(gm[:,2]/gm[:,1])**2)*D

		A, unc_A = D/self.meta['real_time'], unc_D/self.meta['real_time']
		return D, unc_D, A, unc_A

	def _peak(self, x, A, mu, sig, R, alpha, step):
		r2 = 1.41421356237
		return A*np.exp(-0.5*((x-mu)/sig)**2)+R*A*np.exp((x-mu)/(alpha*sig))*erfc((x-mu)/(r2*sig)+1.0/(r2*alpha))+step*A*erfc((x-mu)/(r2*sig))

	def multiplet(self, x, *args):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.fit_config['bg_fit']:
			if self.fit_config['quad_bg']:
				b, peak = 3, args[0]+args[1]*x+args[2]*x**2
			else:
				b, peak = 2, args[0]+args[1]*x
		elif len(x)==len(self._snip):
			b, peak = 0, self._snip.copy()
		else:
			b, peak = 0, self.snip(x)
		R, alpha, step = self.fit_config['R'], self.fit_config['alpha'], self.fit_config['step']
		if self.fit_config['skew_fit']:
			if self.fit_config['step_fit']:
				for n in range(int((len(args)-b)/6)):
					peak += self._peak(x,*args[6*n+b:6*n+6+b])
			else:
				for n in range(int((len(args)-b)/5)):
					peak += self._peak(x,*(args[5*n+b:5*n+5+b]+(step,)))
		else:
			if self.fit_config['step_fit']:
				for n in range(int((len(args)-b)/4)):
					peak += self._peak(x,*(args[4*n+b:4*n+b+3]+(R, alpha)+(args[4*n+3+b])))
			else:
				for n in range(int((len(args)-b)/3)):
					peak += self._peak(x,*(args[3*n+b:3*n+b+3]+(R,alpha,step)))
		return peak

	def _get_p0(self, lh, px, gm, istp):
		ix = {'A':[],'mu':[],'sig':[],'R':[],'alpha':[],'step':[]}
		N = int(self.fit_config['bg_fit'])*(2+int(self.fit_config['quad_bg']))
		if self.fit_config['bg_fit']:
			M = np.column_stack([self.channels[lh[0]:lh[1]]**m for m in range(N)])
			b = np.array([np.sum(self.channels[lh[0]:lh[1]]**m*self._snip[lh[0]:lh[1]]) for m in range(N)])
			M_inv = np.linalg.inv(np.dot(M.T, M))
			p0 = np.dot(M_inv, b).tolist()
			resid = self._snip[lh[0]:lh[1]]-np.dot(M, p0)
			unc = np.sqrt(np.abs(M_inv*np.dot(resid.T, resid)/max([float((lh[1]-lh[0])-N),1.0])))
			bounds = [[i-10*unc[n][n] for n,i in enumerate(p0)],[i+10*unc[n][n] for n,i in enumerate(p0)]]
		else:
			p0, bounds = [],[[],[]]

		R, alpha, step = self.fit_config['R'], self.fit_config['alpha'], self.fit_config['step']
		bA, bm, bs = self.fit_config['A_bound'], self.fit_config['mu_bound'], self.fit_config['sig_bound']
		for n,g in enumerate(gm):
			p0 += px[n]
			bounds[0] += [0.0, p0[-2]-p0[-1]*bm, p0[-1]/bs]
			bounds[1] += [p0[-3]*bA, p0[-2]+p0[-1]*bm, p0[-1]*bs]
			ix['A'].append(N)
			ix['mu'].append(N+1)
			ix['sig'].append(N+2)
			N += 3
			if self.fit_config['skew_fit']:
				p0 += [R, alpha]
				bounds[0] += [0.0, 0.5]
				bounds[1] += [1.0, max((2.5, alpha))]
				ix['R'].append(N)
				ix['alpha'].append(N+1)
				N += 2
			if self.fit_config['step_fit']:
				p0.append(step)
				bounds[0].append(0.0)
				bounds[1].append(0.1)
				ix['step'].append(N)
				N += 1

		if self.fit_config['threads']>1:
			return {'p0':p0, 'lh':lh, 'bounds':bounds, 'gm':gm, 'istp':istp, 'ix':ix, 'x':self.channels[lh[0]:lh[1]],
					 'y':self.hist[lh[0]:lh[1]], 'snip':self.snip(self.channels[lh[0]:lh[1]]), 'fit_config':self.fit_config.copy()}
		else:
			return {'p0':p0, 'lh':lh, 'bounds':bounds, 'gm':gm, 'istp':istp, 'ix':ix}

	def _get_multiplets(self):
		p0_A, cov_p0_A = self._fit_forward_activity()
		pks, p0 = [], {'idx':[],'sig':[],'A':[]}
		W, o = self.fit_config['pk_width'], 0
		for n,gm in enumerate(self.gamma_list):
			if len(gm):
				idx = self.cb.map_idx(gm[:,0])
				sig = self.cb.res(idx)
				A = p0_A[n-o]*self.cb.eff(gm[:,0])*gm[:,1]/sig
				SNR = A/np.sqrt(self._snip[idx])
				cut = np.where((SNR>self.fit_config['SNR_cut'])&(idx>W*sig)&(idx<(len(self._hist)-W*sig)))
				idx, sig, gm_idx, A = idx[cut], sig[cut], np.arange(len(gm))[cut], A[cut]
				p0['idx'].append(idx)
				p0['sig'].append(sig)
				p0['A'].append(A)
				l, h = np.array(idx-W*sig, dtype=np.int32), np.array(idx+W*sig, dtype=np.int32)
				pks.append(np.array([l, h, np.repeat(n, len(l)), gm_idx, np.repeat(o, len(l)), np.arange(len(l)), np.repeat(n, len(l))], dtype=np.int32).T)
			else:
				o += 1
		

		if 'p0' in self.fit_config:
			mp0 = self.fit_config['p0']
			if type(mp0)==dict:
				mp0 = [{i:mp0[i][n] for i in mp0} for n in range(len([mp0[i] for i in mp0][0]))]

			for n,p in enumerate(mp0):
				if 'mu' in p:
					p['mu'] = int(p['mu'])
					if 'E' not in p:
						p['E'] = self.cb.eng(p['mu'])
				elif 'E' in p:
					p['mu'] = self.cb.map_idx(p['E'])
				else:
					raise ValueError('E or mu must be specified in p0.')
				if 'I' not in p:
					p['I'] = -1.0
				if 'unc_I' not in p:
					p['unc_I'] = 1.0
				if 'A' not in p:
					p['A'] = self._hist[p['mu']]-self._snip[p['mu']]
				if 'sig' not in p:
					p['sig'] = self.cb.res(p['mu'])
				if 'istp' not in p:
					p['istp'] = None

				if p['istp'] not in self.meta['istp']:
					self.meta['istp'].append(p['istp'])
					self.gamma_list.append(np.array([[p['E'],p['I'],p['unc_I']]]))
					ig = len(self.gamma_list)-1
				else:
					ig = self.meta['istp'].index(p['istp'])
					self.gamma_list[ig] = np.append(self.gamma_list[ig], [[p['E'],p['I'],p['unc_I']]], axis=0)
				p0['idx'].append([p['mu']])
				p0['sig'].append([p['sig']])
				p0['A'].append([p['A']])
				l, h = p['mu']-W*p['sig'], p['mu']+W*p['sig']
				pks.append(np.array([[l], [h], [len(p0['A'])-1], [len(self.gamma_list[ig])-1], [o], [0], [ig]], dtype=np.int32).T)

		if len(pks)==0:
			return pks

		pks = np.concatenate(pks)
		pks = pks[np.argsort(pks[:,0])]
		pairs = np.where(pks[:-1,1]>pks[1:,0])[0]
		pairs = np.unique(np.concatenate((pairs, pairs+1)))
		if len(pairs):
			groups = [[pairs[0], pairs[1]]]
			for p in pairs[2:]:
				if (pks[p][0]<pks[groups[-1][-1]][1]) and len(groups[-1])<8:
					groups[-1].append(p)
				else:
					groups.append([p])
		pk_pairs = [[pks[i] for i in p] for p in groups] if len(pairs) else []
		pks = [[p] for p in np.delete(pks, pairs, axis=0)]+pk_pairs

		multiplets = []
		for p in pks:
			lh = [p[0][0], p[-1][1]]
			N = [(i[2]-i[4], i[3], i[5], i[6]) for i in p]
			px = [[p0['A'][n][j], p0['idx'][n][j], p0['sig'][n][j]] for n,m,j,k in N]
			gm = np.asarray([self.gamma_list[k][m] for n,m,j,k in N])
			istp = [self.meta['istp'][k] for n,m,j,k in N]
			multiplets.append(self._get_p0(lh, px, gm, istp))

		return multiplets

	def _multi_fit(self, multi):
		p0, lh, bounds, gm = multi['p0'], multi['lh'], multi['bounds'], multi['gm']
		if 'fit' not in multi:
			try:
				fit, unc = curve_fit(self.multiplet, self.channels[lh[0]:lh[1]], 
					self.hist[lh[0]:lh[1]], p0=p0, bounds=bounds, sigma=np.sqrt(self.hist[lh[0]:lh[1]]+0.1))
				converged = True
			except:
				fit, unc = np.array(p0), np.inf*np.ones((len(p0), len(p0)))
				converged = False
		else:
			fit, unc, converged = multi['fit'], multi['unc'], multi['converged']
		
		N, unc_N = self._counts(fit, unc, lh)
		D, unc_D, A, unc_A = self._decays(N, unc_N, gm)
		chi2 = self._chi2(fit, lh)

		cols = ['spec_id','filename','isotope','energy','counts','unc_counts',
				'intensity','unc_intensity','efficiency','unc_efficiency',
				'decays','unc_decays','decay_rate','unc_decay_rate','chi2','converged']

		return [{'fit':fit,'unc':unc,'l':lh[0],'h':lh[1],'gm':gm,'converged':converged,'ix':multi['ix'],'istp':multi['istp']}, 
				pd.DataFrame({'spec_id':self.meta['spec_id'], 'filename':self._fnm, 'isotope':multi['istp'],
								'energy':gm[:,0], 'counts':N, 'unc_counts':unc_N,
								'intensity':gm[:,1], 'unc_intensity':gm[:,2], 'efficiency':self.cb.eff(gm[:,0]),
								'unc_efficiency':self.cb.unc_eff(gm[:,0]), 'decays':D, 'unc_decays':unc_D,
								'decay_rate':A, 'unc_decay_rate':unc_A, 'chi2':chi2, 'converged':converged}, columns=cols)]

	@property
	def fits(self):
		if self._fits is None:
			p0 = self._get_multiplets()
			if len(p0):
				if self.fit_config['threads']>1:
					pool = multiprocessing.Pool(processes=self.fit_config['threads'])
					p0 = pool.map(_parallel_fit, p0)
					pool.close()
					pool.join()
				multiplets = sorted(list(map(self._multi_fit, p0)), key=lambda h:h[0]['l'])
				self._fits = [i[0] for i in multiplets]
				self._peaks = pd.concat([i[1] for i in multiplets], ignore_index=True)
			else:
				self._fits = []
				self._peaks = pd.DataFrame([])
			self._update_db(False)
		return self._fits


	@property
	def peaks(self):
		if self._peaks is None:
			fits = self.fits
		return self._peaks

	def _split_fits(self):
		fits, istp, lh, N_sub = [], [], [], []
		cfg = self.fit_config
		for ft in self.fits:			
			if not ft['converged']:
				N_sub.append(0)
				continue
			f = ft['fit']
			B = int(cfg['bg_fit'])*(2+int(cfg['quad_bg']))
			p0 = f[:B].tolist()
			L = 3+2*int(cfg['skew_fit'])+int(cfg['step_fit'])
			N_sub.append(int((len(f)-B)/L))
			for n in range(N_sub[-1]):
				istp.append(ft['istp'][n])
				fits.append(p0+f[B+n*L:B+(n+1)*L].tolist())
				mu, sig = f[B+n*L+1], f[B+n*L+2]
				lh.append([mu-cfg['pk_width']*sig, mu+cfg['pk_width']*sig])
		itp_set = sorted(list(set(istp)))
		return {i:[[fits[n],lh[n]] for n,ip in enumerate(istp) if ip==i] for i in itp_set}, itp_set, N_sub

	

	def saveas(self, *fnms):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		for fl in fnms:
			if fl.startswith('*'):
				fl = os.path.join(self._path, '.'.join(self._fnm.split('.')[:-1])+'.'+fl.split('.')[-1])

			if any([fl.endswith(e) for e in ['.png','.pdf','.eps','.pgf','.ps','.raw','.rgba','.svg','.svgz']]):
				self.plot(saveas=fl, show=False)

			if fl.endswith('.Spe'):
				### Maestro ASCII .Spe ###
				self._meta['DATE_MEA'] = [dtm.datetime.strftime(self.meta['start_time'], '%m/%d/%Y %H:%M:%S')]
				self._meta['MEAS_TIM'] = ['{0} {1}'.format(int(self.meta['live_time']), int(self.meta['real_time']))]
				self._meta['ENER_FIT'] = ['{0} {1}'.format(self.cb.engcal[0], self.cb.engcal[1])]
				self._meta['MCA_CAL'] = ['3','{0} {1} {2} keV'.format(self.cb.engcal[0], self.cb.engcal[1], (self.cb.engcal[2] if len(self.cb.engcal)>2 else 0.0))]
				defaults = {'ROI':['0'],'SPEC_REM':['DET# 0','DETDESC# None','AP# Maestro Version 7.01'],'PRESETS':['0'],
							'SHAPE_CAL':['3','0E+00 0E+00 0E+00'],'SPEC_ID':['No sample description was entered.']}
				for d in defaults:
					if d not in self.meta:
						self._meta[d] = defaults[d]
				ss = '\n'.join(['${}:\n'.format(sc)+'\n'.join(self.meta[sc]) for sc in ['SPEC_ID','SPEC_REM','DATE_MEA','MEAS_TIM']])
				ss += '\n$DATA:\n0 {}\n'.format(len(self.hist)-1)
				ss += '\n'.join([' '*(8-len(i))+i for i in map(str, self.hist)])+'\n'
				ss += '\n'.join(['${}:\n'.format(sc)+'\n'.join(self.meta[sc]) for sc in ['ROI','PRESETS','ENER_FIT','MCA_CAL','SHAPE_CAL']])+'\n'
				with open(fl,'w') as f:
					f.write(ss)

			if fl.endswith('.Chn'):
				### Maestro integer .Chn ###
				if 'SPEC_REM' in self.meta:
					det_no = int(self.meta['SPEC_REM'][0].split(' ')[1].strip())
				else:
					det_no = 0
				ss = np.array([-1, det_no, 1], dtype='i2').tobytes()
				st = self.meta['start_time']
				ss += np.array(('0' if st.second<10 else '')+str(st.second), dtype='S2').tobytes()
				ss += np.array([self.meta['real_time']/0.02, self.meta['live_time']/0.02], dtype='i4').tobytes()
				months = {1:'Jan',2:'Feb',3:'Mar',4:'Apr',5:'May',6:'Jun',7:'Jul',8:'Aug',9:'Sep',10:'Oct',11:'Nov',12:'Dec'}
				ss += np.array(('0' if st.day<10 else '')+str(st.day)+months[st.month]+str(st.year)[-2:]+('1' if st.year>1999 else '0'), dtype='S8').tobytes()
				ss += np.array(('0' if st.hour<10 else '')+str(st.hour)+('0' if st.minute<10 else '')+str(st.minute), dtype='S4').tobytes()
				ss += np.array([0, len(self.hist)], dtype='i2').tobytes()
				ss += np.array(self.hist, dtype='i4').tobytes()
				ss += np.array([-102, 0], dtype='i2').tobytes()
				ss += np.array(self.cb.engcal, dtype='f4').tobytes()
				if 'SHAPE_CAL' in self.meta:
					ss += np.array(self.meta['SHAPE_CAL'][1].split(' '), dtype='f4').tobytes()
				else:
					ss += np.array([0,0,0], dtype='f4').tobytes()
				ss += np.zeros(228, dtype='i1').tobytes()
				if 'SPEC_REM' in self.meta:
					L = len(self.meta['SPEC_REM'][1].split('# ')[1])
					L = min((63, L))
					ss += np.array(L, dtype='i1').tobytes()
					ss += np.array(self.meta['SPEC_REM'][1].split('# ')[1][:L], dtype='S{}'.format(L)).tobytes()
				else:
					ss += np.array(0, dtype='i1').tobytes()
				if 'SPEC_ID' in self.meta:
					if self.meta['SPEC_ID'][0]!='No sample description was entered.':
						L = len(''.join(self.meta['SPEC_ID']))
						L = min((63, L))
						ss += np.array(L, dtype='i1').tobytes()
						ss += np.array(''.join(self.meta['SPEC_ID'])[:L])
					else:
						ss += np.array(0, dtype='i1').tobytes()
						ss += np.array('\x00'*63, dtype='S63').tobytes()
				else:
					ss += np.array(0, dtype='i1').tobytes()
					ss += np.array('\x00'*63, dtype='S63').tobytes()
				ss += np.array('\x00'*128, dtype='S128').tobytes()
				with open(fl, 'wb') as f:
					f.write(ss)

			if fl.endswith('.spe'):
				### Radware gf3 .spe ###
				name = self._fnm[:8]+' '*(8-len(self._fnm))
				with open(fl, 'wb') as f:
					ss = np.array(24, dtype=np.uint32).tobytes()
					ss += np.array(name, dtype='c').tobytes()
					ss += np.array([len(self.hist), 1, 1, 1, 24, 4*len(self.hist)], dtype=np.uint32).tobytes()
					ss += np.array(self.hist, dtype=np.float32).tobytes()
					ss += np.array(4*len(self.hist), dtype=np.uint32).tobytes()
					f.write(ss)

			if fl.endswith('.csv'):
				self.peaks.to_csv(fl, index=False)

			if fl.endswith('.db'):
				meta = self.meta
				self._check_db(fl)
				self.meta = {i:meta[i] for i in meta if i.endswith('cal')}
				self._update_db(True)
				self._update_db(False)


	def _from_Spe(self, filename=None):
		if filename is None:
			filename = os.path.join(self._path, self._fnm)
		with open(filename) as f:
			ln = f.readline()
			while ln:
				if ln.startswith('$'):
					section = ln.strip()[1:-1]
					if section=='DATA':
						L = int(f.readline().strip().split(' ')[1])+1
						self.hist = np.fromfile(f, dtype=np.int64, count=L, sep='\n')
					else:
						self._meta[section] = []
				else:
					self._meta[section].append(ln.strip())
				ln = f.readline()

		self._meta['start_time'] = dtm.datetime.strptime(self._meta['DATE_MEA'][0], '%m/%d/%Y %H:%M:%S')
		self._meta['real_time'], self._meta['live_time'] = tuple(map(float, self._meta['MEAS_TIM'][0].split(' ')))
		self._meta['engcal'] = list(map(float, self._meta['MCA_CAL'][-1].split(' ')[:-1]))


	def _from_Chn(self, filename=None):
		if filename is None:
			filename = os.path.join(self._path, self._fnm)
		with open(filename, 'rb') as f:
			det_no = np.frombuffer(f.read(6), dtype='i2')[1]
			sts = np.frombuffer(f.read(2), dtype='S2')[0].decode('utf-8')
			self._meta['real_time'], self._meta['live_time'] = tuple(map(float, 0.02*np.frombuffer(f.read(8), dtype='i4')))
			st = np.frombuffer(f.read(12), dtype='S12')[0].decode('utf-8')
			months = {'jan':'01','feb':'02','mar':'03','apr':'04',
						'may':'05','jun':'06','jul':'07','aug':'08',
						'sep':'09','oct':'10','nov':'11','dec':'12'}
			start_time = '{0}/{1}/{2} {3}:{4}:{5}'.format(months[st[2:5].lower()],st[:2],('20' if st[7]=='1' else '19')+st[5:7],st[8:10],st[10:],sts)
			self._meta['start_time'] = dtm.datetime.strptime(start_time, '%m/%d/%Y %H:%M:%S')
			L = np.frombuffer(f.read(4), dtype='i2')[1]
			self.hist = np.asarray(np.frombuffer(f.read(4*L), dtype='i4'), dtype=np.int64)
			f.read(4)
			self._meta['engcal'] = np.frombuffer(f.read(12),dtype='f4').tolist()

			shape = np.frombuffer(f.read(12), dtype='f4').tolist()
			self._meta['SHAPE_CAL'] = ['3', ' '.join(map(str, shape))]
			f.read(228)
			L = np.frombuffer(f.read(1), dtype='i1')[0]
			if L:
				det_desc = np.frombuffer(f.read(L), dtype='S{}'.format(L))[0].decode('utf-8')
				self._meta['SPEC_REM'] = ['DET# '+str(det_no), 'DETDESC# '+det_desc, 'AP# Maestro Version 7.01']
			if L<63:
				f.read(63-L)
			L = np.frombuffer(f.read(1), dtype='i1')[0]
			if L:
				sample_desc = np.frombuffer(f.read(L),dtype='S{}'.format(L))[0].decode('utf-8')
				self._meta['SPEC_ID'] = [sample_desc]

	def summarize(self):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		print(self.__str__())

	def __str__(self):
		ss = ''
		for n,p in self.peaks.iterrows():
			ln1 = [p['energy'], p['isotope'], p['intensity']*100.0]
			ln0 = '{1} - {0} keV (I = {2}%)'.format(*ln1)
			ss += ln0 + '\n'
			ss += ''.join(['-']*len(ln0)) + '\n'
			ln2 = [int(p['counts']), (int(p['unc_counts']) if np.isfinite(p['unc_counts']) else np.inf),
					format(p['decays'], '.3e'), format(p['unc_decays'], '.3e'), 
					format(p['decay_rate'], '.3e'), format(p['unc_decay_rate'], '.3e'), 
					format(p['decay_rate']/3.7E4, '.3e'), format(p['unc_decay_rate']/3.7E4, '.3e'), 
					round(p['chi2'],3)]
			ss += 'counts: {0} +/- {1}'.format(ln2[0], ln2[1]) + '\n'
			ss += 'decays: {0} +/- {1}'.format(ln2[2], ln2[3]) + '\n'
			ss += 'activity (Bq): {0} +/- {1}'.format(ln2[4], ln2[5]) + '\n'
			ss += 'activity (uCi): {0} +/- {1}'.format(ln2[6], ln2[7]) + '\n'
			ss += 'chi2/dof: {}'.format(ln2[8]) + '\n'
			ss += '\n'
		return ss


	def plot(self, fit=True, labels=True, snip=False, xcalib=True, **kwargs):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		
		xgrid = np.array([self.channels-0.5, self.channels+0.5]).T.flatten()
		spec = np.array([self.hist, self.hist]).T.flatten()
		erange = self.cb.eng(xgrid) if xcalib else xgrid

		f, ax = _init_plot(figsize=(12.8, 4.8), **kwargs)
		spec_label = self._fnm if self._fnm is not None else 'Spectrum'
		ax.plot(erange, spec, lw=1.2, zorder=1, label=(spec_label if labels else None))

		if snip:
			erange = self.cb.eng(self.channels) if xcalib else self.channels
			ax.plot(erange, self._snip)

		if fit:
			cm, cl = colors(), colors(aslist=True)
			cl = [c for c in cl if c not in [cm['k'], cm['gy']]]
			ls = ['-','-.','--']
			sub, itp, N_sub = self._split_fits()
			
			for n,p in enumerate(self.fits):
				if not p['converged']:
					xgrid = np.arange(p['l'], p['h'], 0.1)
					pk_fit = self.multiplet(xgrid, *p['fit'])
					erange = self.cb.eng(xgrid) if xcalib else xgrid
					ax.plot(erange, np.where(pk_fit>0.1, pk_fit, 0.1), ls=':', lw=1.2, color=cm['gy'])
				elif N_sub[n]>1:
					xgrid = np.arange(p['l'], p['h'], 0.1)
					pk_fit = self.multiplet(xgrid, *p['fit'])
					erange = self.cb.eng(xgrid) if xcalib else xgrid
					ax.plot(erange, np.where(pk_fit>0.1, pk_fit, 0.1), ls='--', lw=1.4, color=cm['gy'])

			lbs = []
			for n,i in enumerate(itp):
				c = cl[n%len(cl)]
				ilbl = Isotope(i).TeX
				for p,lh in sub[i]:
					xgrid = np.arange(lh[0], lh[1], 0.1)
					pk_fit = self.multiplet(xgrid, *p)
					lb = ilbl if (labels and ilbl not in lbs and len(lbs)<20) else None
					erange = self.cb.eng(xgrid) if xcalib else xgrid
					ax.plot(erange, np.where(pk_fit>0.1, pk_fit, 0.1), lw=1.4, color=c, label=lb, ls=ls[int(n/len(cl))%len(ls)])
					if lb is not None:
						lbs.append(lb)
		
		if xcalib:
			ax.set_xlabel('Energy (keV)')
		else:
			ax.set_xlabel('ADC Channel')
		ax.set_ylabel('Counts')
		if labels:
			ncol = min([max([int(len(lbs)/5),1]), 3])
			ax.legend(loc=0, ncol=ncol)

		return _close_plot(f, ax, **kwargs)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import matplotlib.pyplot as plt
from cycler import cycler


def colors(style='default', shade='dark', aslist=False):
	"""Some nice color maps.

	...

	Parameters
	----------

	Returns
	-------

	Notes
	-----

	References
	----------

	Examples
	--------

	"""
	default = {'light':{'k':'#34495e','gy':'#95a5a6','r':'#e74c3c','b':'#3498db','g':'#2ecc71',
	'aq':'#1abc9c','o':'#e67e22','y':'#f1c40f','p':'#9b59b6','w':'#ecf0f1'},
	'dark':{'k':'#2c3e50','gy':'#7f8c8d','r':'#c0392b','b':'#2980b9','g':'#27ae60',
	'aq':'#16a085','o':'#d35400','y':'#f39c12','p':'#8e44ad','w':'#bdc3c7'}}

	american = {'light':{'k':'#636e72','gy':'#dfe6e9','r':'#ff7675','b':'#74b9ff','g':'#55efc4',
	'aq':'#81ecec','o':'#fab1a0','y':'#ffeaa7','p':'#a29bfe','w':'#ecf0f1'},
	'dark':{'k':'#2d3436','gy':'#b2bec3','r':'#d63031','b':'#0984e3','g':'#00b894',
	'aq':'#00cec9','o':'#e17055','y':'#fdcb6e','p':'#6c5ce7','w':'#bdc3c7'}}

	aussie = {'light':{'k':'#95afc0','gy':'#dff9fb','r':'#ff7979','b':'#686de0','g':'#badc58',
	'aq':'#7ed6df','o':'#ffbe76','y':'#f6e58d','p':'#e056fd','w':'#ecf0f1'},
	'dark':{'k':'#535c68','gy':'#c7ecee','r':'#eb4d4b','b':'#4834d4','g':'#6ab04c',
	'aq':'#22a6b3','o':'#f0932b','y':'#f9ca24','p':'#be2edd','w':'#bdc3c7'}}

	british = {'light':{'k':'#353b48','gy':'#7f8fa6','r':'#e84118','b':'#273c75','g':'#4cd137',
	'aq':'#1abc9c','o':'#e67e22','y':'#fbc531','p':'#9c88ff','w':'#f5f6fa'},
	'dark':{'k':'#2f3640','gy':'#718093','r':'#c23616','b':'#192a56','g':'#44bd32',
	'aq':'#16a085','o':'#d35400','y':'#e1b12c','p':'#8c7ae6','w':'#dcdde1'}}

	canadian = {'light':{'k':'#576574','gy':'#c8d6e5','r':'#ff6b6b','b':'#54a0ff','g':'#1dd1a1',
	'aq':'#00d2d3','o':'#feca57','y':'#f1c40f','p':'#5f27cd','w':'#ecf0f1'},
	'dark':{'k':'#222f3e','gy':'#8395a7','r':'#ee5253','b':'#2e86de','g':'#10ac84',
	'aq':'#01a3a4','o':'#ff9f43','y':'#f39c12','p':'#341f97','w':'#bdc3c7'}}

	chinese = {'light':{'k':'#57606f','gy':'#a4b0be','r':'#ff6b81','b':'#70a1ff','g':'#7bed9f',
	'aq':'#1abc9c','o':'#ff7f50','y':'#eccc68','p':'#5352ed','w':'#ffffff'},
	'dark':{'k':'#2f3542','gy':'#747d8c','r':'#ff4757','b':'#1e90ff','g':'#2ed573',
	'aq':'#16a085','o':'#ff6348','y':'#ffa502','p':'#3742fa','w':'#f1f2f6'}}

	german = {'light':{'k':'#778ca3','gy':'#d1d8e0','r':'#fc5c65','b':'#45aaf2','g':'#26de81',
	'aq':'#2bcbba','o':'#fd9644','y':'#fed330','p':'#a55eea','w':'#ecf0f1'},
	'dark':{'k':'#4b6584','gy':'#a5b1c2','r':'#eb3b5a','b':'#2d98da','g':'#20bf6b',
	'aq':'#0fb9b1','o':'#fa8231','y':'#f7b731','p':'#8854d0','w':'#bdc3c7'}}

	spanish = {'light':{'k':'#34495e','gy':'#d1ccc0','r':'#ff5252','b':'#34ace0','g':'#33d9b2',
	'aq':'#1abc9c','o':'#ff793f','y':'#ffb142','p':'#706fd3','w':'#f7f1e3'},
	'dark':{'k':'#2c3e50','gy':'#84817a','r':'#b33939','b':'#227093','g':'#218c74',
	'aq':'#16a085','o':'#cd6133','y':'#cc8e35','p':'#474787','w':'#aaa69d'}}

	swedish = {'light':{'k':'#485460','gy':'#d2dae2','r':'#ff5e57','b':'#4bcffa','g':'#0be881',
	'aq':'#34e7e4','o':'#ffc048','y':'#ffdd59','p':'#575fcf','w':'#ecf0f1'},
	'dark':{'k':'#1e272e','gy':'#808e9b','r':'#ff3f34','b':'#0fbcf9','g':'#05c46b',
	'aq':'#00d8d6','o':'#ffa801','y':'#ffd32a','p':'#3c40c6','w':'#bdc3c7'}}

	cm = {'default':default,'american':american,'aussie':aussie,'british':british,'canadian':canadian,'chinese':chinese,'german':german,'spanish':spanish,'swedish':swedish}
	if aslist:
		return [cm[style][shade][c] for c in ['k','gy','r','b','g','o','aq','y','p']]
	return cm[style][shade]


def set_style(sty='show'):
	"""Preset styles for various purposes.

	...

	Parameters
	----------

	Returns
	-------

	"""
	cm = colors()
	plt.rcParams['font.family']='sans-serif'
	plt.rcParams['axes.prop_cycle'] = cycler(color=[cm['k'],cm['r'],cm['b'],cm['gy'],cm['g'],cm['p'],cm['o'],cm['aq'],cm['y']])
	plt.rcParams['figure.autolayout'] = 'True'
	plt.rcParams['xtick.minor.visible']='True'
	plt.rcParams['ytick.minor.visible']='True'
	plt.rcParams['axes.titlepad']='8'
	if sty=='show':
		plt.rcParams['font.size']='14'
		plt.rcParams['xtick.major.pad']='8'
		plt.rcParams['ytick.major.pad']='8'
		plt.rcParams['legend.fontsize']='12'
		plt.rcParams['lines.markersize']='4.5'
		plt.rcParams['errorbar.capsize']='5.0'
		plt.rcParams['lines.linewidth'] = '1.8'
	elif sty=='paper':
		plt.rcParams['font.size']='14'
		plt.rcParams['xtick.major.pad']='8'
		plt.rcParams['ytick.major.pad']='8'
		plt.rcParams['legend.fontsize']='12'
		plt.rcParams['lines.markersize']='3.5'
		plt.rcParams['errorbar.capsize']='4.0'
		plt.rcParams['lines.linewidth'] = '2.4'
	elif sty=='poster':
		plt.rcParams['font.size']='18'
		plt.rcParams['font.weight']='bold'
		plt.rcParams['axes.titleweight']='bold'
		plt.rcParams['axes.titlepad']='14'
		plt.rcParams['xtick.major.pad']='6'
		plt.rcParams['ytick.major.pad']='6'
		plt.rcParams['xtick.major.size']='5'
		plt.rcParams['xtick.minor.size']='3.5'
		plt.rcParams['xtick.major.width']='1.5'
		plt.rcParams['xtick.minor.width']='1.2'
		plt.rcParams['ytick.major.size']='5'
		plt.rcParams['ytick.minor.size']='3.5'
		plt.rcParams['ytick.major.width']='1.5'
		plt.rcParams['ytick.minor.width']='1.2'
		plt.rcParams['axes.linewidth']='1.2'
		plt.rcParams['legend.fontsize']='14'
		plt.rcParams['lines.markersize']='4.5'
		plt.rcParams['errorbar.capsize']='5.0'
		plt.rcParams['lines.linewidth'] = '2.8'
	elif sty=='presentation':
		plt.rcParams['font.size']='14'
		plt.rcParams['xtick.major.pad']='8'
		plt.rcParams['ytick.major.pad']='8'
		plt.rcParams['legend.fontsize']='12'
		plt.rcParams['lines.markersize']='3.5'
		plt.rcParams['errorbar.capsize']='4.0'
		plt.rcParams['lines.linewidth'] = '1.8'


def _init_plot(**kwargs):
	f, ax = None, None
	if 'f' in kwargs and 'ax' in kwargs:
		f, ax = kwargs['f'], kwargs['ax']
	
	N = kwargs['N_plots'] if 'N_plots' in kwargs else 1

	if f is None or ax is None:
		if 'figsize' in kwargs:
			f, ax = plt.subplots(1, N, figsize=kwargs['figsize'])
		else:
			f, ax = plt.subplots(1, N)

	if 'style' in kwargs:
		set_style(kwargs['style'])

	return f, ax


def _close_plot(fig, axis, **kwargs):
	f, ax = fig, axis

	if 'default_log' in kwargs:
		if kwargs['default_log']:
			ax.set_yscale('log')
	else:
		ax.set_yscale('log')

	if 'scale' in kwargs:
		s = kwargs['scale'].lower()
		if s in ['log','logy']:
			ax.set_yscale('log')
		elif s in ['lin','liny']:
			ax.set_yscale('linear')
		elif s=='logx':
			ax.set_xscale('log')
		elif s=='linx':
			ax.set_xscale('linear')
		elif s in ['linlog','loglog','loglin','linlin']:
			ax.set_yscale(s[3:].replace('lin','linear'))
			ax.set_xscale(s[:3].replace('lin','linear'))

	if 'logscale' in kwargs:
		if kwargs['logscale']:
			ax.set_yscale('log')
		else:
			ax.set_yscale('linear')

	if 'logx' in kwargs:
		if kwargs['logx']:
			ax.set_xscale('log')
		else:
			ax.set_xscale('linear')

	if 'logy' in kwargs:
		if kwargs['logy']:
			ax.set_yscale('log')
		else:
			ax.set_yscale('linear')

	f.tight_layout()

	if 'saveas' in kwargs:
		if kwargs['saveas'] is not None:
			f.savefig(kwargs['saveas'])

	if 'show' in kwargs:
		if kwargs['show']:
			plt.show()
	else:
		plt.show()

	if 'f' in kwargs and 'ax' in kwargs:
		return f, ax

	plt.close()

set_style()



"""
npat
=====

Provides
  1. An array object of arbitrary homogeneous items
  2. Fast mathematical operations over arrays
  3. Linear Algebra, Fourier Transforms, Random Number Generation

How to use the documentation
----------------------------
Documentation is available in two forms: docstrings provided
with the code, and a loose standing reference guide, available from
`the NumPy homepage <https://www.scipy.org>`_.

Code snippets are indicated by three greater-than signs::

  >>> x = 42
  >>> x = x + 1

Use the built-in ``help`` function to view a function's docstring::

  >>> help(np.sort)
  ... # doctest: +SKIP


Available subpackages
---------------------
doc
    Topical documentation on broadcasting, indexing, etc.
lib
    Basic functions used by several sub-packages.
random
    Core Random Tools
linalg
    Core Linear Algebra Tools
fft
    Core FFT routines
polynomial
    Polynomial tools
testing
    NumPy testing tools
f2py
    Fortran to Python Interface Generator.
distutils
    Enhancements to distutils with support for
    Fortran compilers support and more.

Utilities
---------
test
    Run numpy unittests
show_config
    Show numpy build configuration
dual
    Overwrite certain functions with high-performance Scipy tools
matlib
    Make everything matrices.
__version__
    NumPy version string


"""

from .dbmgr import *
from .spectroscopy import *
from .isotope import *
from .decay_chain import *
from .plotter import *
from .reaction import *
from .irradiation import *
from .listfiles import *

name = "npat"
__version__ = '0.3.5'
__all__ = ['get_cursor', 'get_connection', 'colors',
			'set_style', 'Spectrum', 'Calibration', 
			'Isotope', 'DecayChain', 'Reaction',
			'Ziegler', 'Library', 'Element', 
			'download', 'MVME']


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import os
import numpy as np
import sqlite3
import pandas as pd

from .plotter import colors
from .plotter import _init_plot
from .plotter import _close_plot
from .dbmgr import get_cursor
from .isotope import Isotope

global ZG_CONNECTIONS_DICT
ZG_CONNECTIONS_DICT = {}

#### TODO: allow new compound definitions as {'El':wt_pct} and accept .json stack

class Ziegler(object):
	"""Method for solving energy loss within stacked-target foil experiment

	...

	Parameters
	----------
	x : type
		Description of parameter `x`.

	Attributes
	----------

	Methods
	-------

	Notes
	-----

	References
	----------

	Examples
	--------

	"""

	def __init__(self, stack, beam=None, **kwargs):
		### stack is list of dicts, which must have 'compound' specified.
		### Areal density specified by either 'ad' (mg/cm^2), both mass (g) and area (cm^2),
		### 'thickness' (mm) or both 'density' (g/cm^3) and 'thickness' (mm)

		zdb = get_cursor('ziegler')
		
		self.protons = {int(i[0]):list(map(float,i[1:])) for i in zdb.execute('SELECT * FROM protons')}
		self.helium = {int(i[0]):list(map(float,i[1:])) for i in zdb.execute('SELECT * FROM helium')}
		self.ionization = {int(i[0]):list(map(float,i[1:])) for i in zdb.execute('SELECT * FROM ionization')}
		self.weights = {int(i[0]):list(map(float,i[1:3])) for i in zdb.execute('SELECT * FROM weights')}
		self.compounds = {str(i[0]):[[int(h.split(':')[0]), float(h.split(':')[1])] for h in i[2].split(',')] for i in zdb.execute('SELECT * FROM compounds')}
		self.compounds = {cm:[[i[0],i[1]/sum([m[1] for m in self.compounds[cm]])] for i in self.compounds[cm]] for cm in self.compounds}
		self.densities = {str(i[0]):float(i[1]) for i in zdb.execute('SELECT * FROM compounds')}
		self.elements = sorted([[str(i[0]), int(i[2].split(':')[0])] for i in zdb.execute('SELECT * FROM compounds') if len(i[2].split(':'))==2], key=lambda h:len(h[0]), reverse=True)

		self._meta = {}
		self.meta = {'beam_istp':'1H', 'E0':33.0, 'dE0':0.3, 'N':10000, 'dp':1.0,
						'chunk_size':1E7, 'threads':1,
						'solved':False, 'accuracy':0.01, 'min_steps':2, 'max_steps':50}
		self._stack = []
		if beam is not None:
			print('Keyword `beam` deprecated: see documentation for proper input.')
			self.meta = beam
		self.meta = kwargs
		if type(stack)==str:
			stack = pd.read_csv(stack)
			stack = [{i:r[i] for i in stack.columns if not (np.isnan(r[i]) if type(r[i])==float else False)} for n,r in stack.iterrows()]
		self.stack = stack

	def check_db(self, db=None):
		if db is not None:
			path, fnm = os.path.split(db)
			if path in ['',' ']:
				path = os.getcwd()
			if fnm in ['',' ']:
				raise ValueError('Invalid db Filename: {}'.format(db))
			db_fnm = os.path.join(path, fnm)

			global ZG_CONNECTIONS_DICT
			if os.path.exists(db_fnm):
				if db_fnm not in ZG_CONNECTIONS_DICT:
					ZG_CONNECTIONS_DICT[db_fnm] = sqlite3.connect(db_fnm)
				self.db_connection = ZG_CONNECTIONS_DICT[db_fnm]
				self.db = self.db_connection.cursor()
			else:
				print('WARNING: DB {} does not exist, creating new file.'.format(fnm))
				from sqlite3 import Error
				try:
					self.db_connection = sqlite3.connect(db_fnm)
					ZG_CONNECTIONS_DICT[db_fnm] = self.db_connection
					self.db = self.db_connection.cursor()
				except Error as e:
					print(e)

	@property
	def meta(self):
		return self._meta

	@meta.setter
	def meta(self, meta_dict):
		for nm in meta_dict:
			self._meta[nm] = meta_dict[nm]
			if nm in ['beam_istp', 'istp']:
				_ITP = Isotope(meta_dict[nm])
				self._meta['Z'] = _ITP.Z
				self._meta['amu'] = _ITP.mass
		if self._meta['min_steps']>self._meta['max_steps']:
			self._meta['max_steps'] = self._meta['min_steps']+1
			print('WARNING: min_steps > max_steps, setting max_steps to {}'.format(self._meta['max_steps']))
		self._meta['solved'] = False

	def __getitem__(self, key):
		if type(key)==str:
			for s in self.stack:
				if s['name']==key:
					return s
			return None
		else:
			return self.stack[int(key)]

	@property
	def stack(self):
		if not self.meta['solved']:
			self._solve()
		return self._stack

	@stack.setter
	def stack(self, _stack):
		self._stack = list(_stack)
		self._meta['solved'] = False
		for s in self._stack:
			if 'name' not in s:
				s['name'] = None
			if 'compound' not in s:
				raise ValueError('compound must be specified')

			if type(s['compound'])==dict:
				cs = ''
				for c in s['compound']:
					print(c)
					cs = c
					self.compounds[c] = s['compound'][c]
					print(self.compounds[c])
				s['compound'] = cs
				if s['compound']=='':
					raise ValueError('compound must be specified')

			if 'ad' not in s:
				if 'area' in s and 'mass' in s:
					s['ad'] = 1e3*s['mass']/s['area']
				elif 'density' in s and 'thickness' in s:
					s['ad'] = 100.0*s['density']*s['thickness']
				elif s['compound'] in self.densities and 'thickness' in s:
					s['ad'], s['density'] = 100.0*self.densities[s['compound']]*s['thickness'], self.densities[s['compound']]

			if 'density' not in s:
				if s['compound'] in self.densities:
					s['density'] = self.densities[s['compound']]
					if 'thickness' not in s and 'ad' in s:
						s['thickness'] = s['ad']/(100.0*s['density'])

			if 'ad' not in s:
				raise ValueError('Areal density either not specified or not computable: {}'.format(s))

			if s['compound'] not in self.compounds:
				cm = s['compound']
				nums = [str(i) for i in range(10)]
				cm_list = []
				for el,Z in self.elements:
					if len(cm)==0:
						break
					f = cm.split(el)
					if len(f)>1:
						c = f[0]
						for e in f[1:]:
							if len(e):
								if e[0] in nums:
									cm_list.append([Z, float(e[0])])
									if len(e)>1:
										c += e[1:]
								else:
									cm_list.append([Z, 1.0])
									c += e
							else:
								cm_list.append([Z, 1.0])
						cm = c
				if len(cm)==0 and len(cm_list):
					self.compounds[s['compound']] = [[i[0], i[1]/sum([i[1] for i in cm_list])] for i in cm_list] 

			if s['compound'] not in self.compounds:
				raise ValueError('compound {} not known.'.format(s['compound']))


	def get_S(self, E, cm):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		# energy E in MeV , stopping power in MeV/(mg/cm2)
		E, r0 = np.asarray(E), False
		if not E.shape:
			E, r0 = np.array([E]), True
		S = np.zeros(len(E))
		A_ave = sum([self.weights[z2][0]*w for z2,w in self.compounds[cm]])
		for z2, w in self.compounds[cm]:
			S_nucl = self.get_S_nucl(E, self.meta['Z'], self.meta['amu'], z2, self.weights[z2][0])
			if self.meta['Z']==1:
				S += w*(S_nucl+self.get_S_p(E, z2, self.meta['amu']))
			elif self.meta['Z']==2:
				S += w*(S_nucl+self.get_S_He(E, z2, self.meta['amu']))
			else:
				S += w*(S_nucl+self.get_S_elec(E, z2, self.meta['amu'], self.meta['Z']))
		return S[0]*0.6022140857/A_ave if r0 else S*0.6022140857/A_ave

	def get_S_nucl(self, E, z1, m1, z2, m2):
		RM = (m1+m2)*np.sqrt((z1**(2/3.0)+z2**(2/3.0)))
		ER = 32.53*m2*1E3*E/(z1*z2*RM)
		return (0.5*np.log(1.0+ER)/(ER+0.10718+ER**0.37544))*8.462*z1*z2*m1/RM

	def get_S_p(self, eng, z2, M1=1.00727647):
		S = np.zeros(len(eng))
		E = 1E3*eng/M1
		A = self.protons[z2]
		beta_sq = np.where(E>=1E3,1.0-1.0/(1.0+E/931478.0)**2,0.9)
		B0 = np.where(E>=1E3,np.log(A[6]*beta_sq/(1.0-beta_sq))-beta_sq,0.0)
		Y = np.log(E[(1E3<=E)&(E<=5E4)])
		B0[np.nonzero(np.where((1E3<=E)&(E<=5E4),B0,0))] -= A[7]+A[8]*Y+A[9]*Y**2+A[10]*Y**3+A[11]*Y**4
		S[E>=1E3] = (A[5]/beta_sq[E>=1E3])*B0[E>=1E3]
		S_low = A[1]*E[(10<=E)&(E<1E3)]**0.45
		S_high = (A[2]/E[(10<=E)&(E<1E3)])*np.log(1.0+(A[3]/E[(10<=E)&(E<1E3)])+A[4]*E[(10<=E)&(E<1E3)])
		S[(10<=E)&(E<1E3)] = S_low*S_high/(S_low+S_high)
		S[(0<E)&(E<10)] = A[0]*E[(0<E)&(E<10)]**0.5
		return S

	def get_S_He(self,eng,z2,M1=4.003):
		S = np.zeros(len(eng))
		E = eng*4.0015/M1
		E = np.where(E>=0.001,E,0.001)
		A = self.helium[z2]
		S_low = A[0]*(1E3*E[E<=10])**A[1]
		S_high = (A[2]/E[E<=10])*np.log(1.0+(A[3]/E[E<=10])+A[4]*E[E<=10])
		S[E<=10] = S_low*S_high/(S_low+S_high)
		Y = np.log(1.0/E[E>10])
		S[E>10] = np.exp(A[5]+A[6]*Y+A[7]*Y**2+A[8]*Y**3)
		return S

	def get_S_elec(self, eng, z2, M1, z1):
		S = np.zeros(len(eng))
		E_keV = 1E3*eng
		S[E_keV/M1<1000] = self.get_eff_Z_ratio(E_keV[E_keV/M1<1000],z1,M1)**2*self.get_S_p(eng[E_keV/M1<1000],z2,M1)
		Y = E_keV[E_keV/M1>=1000]/M1
		beta_sq = 1.0-1.0/(1.0+Y/931478.0)**2
		FX = np.log(2E6*0.511003*beta_sq/(1.0-beta_sq))-beta_sq
		ZHY = 1.0-np.exp(-0.2*np.sqrt(Y)-0.0012*Y-0.00001443*Y**2)
		Z1EFF = self.get_eff_Z_ratio(E_keV[E_keV/M1>=1000],z1,M1)*ZHY
		S[E_keV/M1>=1000] = 4E-1*np.pi*(1.9732857/137.03604)**2*Z1EFF**2*z2*(FX-np.log(self.ionization[z2][0]))/(0.511003*beta_sq)
		return S

	def get_eff_Z_ratio(self, E_keV, z1, M1):
		if z1==1:
			return np.ones(len(eng))
		elif z1==2:
			Y = np.log(E_keV/M1)
			return z1*(1.0-np.exp(-0.7446-0.1429*Y-0.01562*Y**2+0.00267*Y**3-0.000001325*Y**8))
		elif z1==3:
			Y = E_keV/M1
			return z1*(1.0-np.exp(-0.7138-0.002797*Y-0.000001348*Y**2))
		BB = -0.886*np.sqrt(0.04*E_keV/M1)/z1**(2/3.0)
		return z1*(1.0-np.exp(BB-0.0378*np.sin(0.5*np.pi*BB))*(1.034-0.1777*np.exp(-0.08114*z1)))

	def _calc_bins(self):
		return np.arange(0.0, self.meta['E0']+10.0*self.meta['dE0'], min([0.1, self.meta['E0']/500.0]))

	def _solve_chunk(self, N):
		E0 = self.meta['E0']+self.meta['dE0']*np.random.normal(size=int(N))
		bins = self._calc_bins()
		hists = []
		dp = self.meta['dp']
		for n, sm in enumerate(self._stack):
			E_bar = [E0]
			if np.average(E0)<=0.0:
				hists.append(np.concatenate([[N],np.zeros(len(bins)-2)]))
			else:
				steps = int((1.0/self.meta['accuracy'])*sm['ad']*dp*self.get_S(np.average(E0), sm['compound'])/np.average(E0))
				steps = min([max([self.meta['min_steps'], steps]), self.meta['max_steps']])
				dr = (1.0/float(steps))
				for i in range(steps):
					S1 = self.get_S(E0, sm['compound'])
					E1 = E0 - dr*dp*sm['ad']*S1
					E1 = np.where(E1>0, E1, 0.0)
					E0 = E0 - dr*0.5*dp*sm['ad']*(S1+self.get_S(E1, sm['compound']))
					E0 = np.where(E0>0, E0, 0.0)
					E_bar.append(E0)
				hists.append(np.histogram(np.concatenate(E_bar), bins=bins)[0])
		return hists

	def _solve(self):
		if self.meta['solved']:
			return
		self.meta['solved'] = True
		dN = np.linspace(0, self.meta['N'], int(np.ceil(self.meta['N']/float(self.meta['chunk_size'])))+1, dtype=int)
		histos = list(map(self._solve_chunk, dN[1:]-dN[:-1]))

		bins = self._calc_bins()
		energy = 0.5*(bins[1:]+bins[:-1])
		for n,sm in enumerate(self._stack):
			sm['flux'] = np.sum([h[n] for h in histos], axis=0)
			sm['flux'] = sm['flux']/np.sum(sm['flux'])
			sm['mu_E'] = np.sum(sm['flux']*energy)
			sm['sig_E'] = np.sqrt(np.sum(sm['flux']*(energy-sm['mu_E'])**2))
			lh = np.where(sm['flux']>0)[0]
			if lh.size:
				if lh[0]==0:
					nm = sm['name'] if sm['name'] is not None else sm['compound']+str(n+1)
					print('WARNING: Beam stopped in foil {}'.format(nm))
			sm['flux'] = sm['flux'][lh[0]:lh[-1]]
			sm['energy'] = energy[lh[0]:lh[-1]]
			sm['bins'] = bins[lh[0]:lh[-1]+1]

	def saveas(self, *fnms):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		cols = ['name','compound','thickness','density','ad','mu_E','sig_E']
		stack = pd.DataFrame([{c:(sm[c] if c in sm else None) for c in cols} for sm in self.stack], columns=cols)
		cols = ['name','energy','flux']
		fluxes = pd.concat([pd.DataFrame({c:sm[c] for c in cols}, columns=cols) for sm in self.stack if sm['name'] is not None], ignore_index=True)
		for fl in fnms:
			if any([fl.endswith(e) for e in ['.png','.pdf','.eps','.pgf','.ps','.raw','.rgba','.svg','.svgz']]):
				self.plot(saveas=fl, show=False)
			if fl.endswith('.csv'):
				stack.to_csv(fl.replace('.csv','_stack.csv'), index=False)
				fluxes.to_csv(fl.replace('.csv','_fluxes.csv'), index=False)
			if fl.endswith('.db'):
				self.check_db(fl)
				stack.to_sql('stack', self.db_connection, if_exists='replace', index=False)
				fluxes.to_sql('fluxes', self.db_connection, if_exists='replace', index=False)

	def summarize(self, samples=None):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		for n,sm in enumerate(self.stack):
			if sm['name'] is not None:
				if samples is not None:
					if not any([re.match(s, sm['name']) for s in samples]):
						continue
				nm = sm['name'] if sm['name'] is not None else sm['compound']+str(n+1)
				print(nm+': '+str(round(sm['mu_E'], 2))+' +/- '+str(round(sm['sig_E'], 2))+' (MeV)')

	def plot_S(self, compound, energy=None, **kwargs):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if energy is None:
			energy = 10.0**np.arange(0.1,2.8,0.1)
		f,ax = _init_plot(**kwargs)
		ax.plot(energy, self.get_S(energy, compound), label=compound.title())
		ax.set_xlabel('Energy (MeV)')
		ax.set_ylabel(r'Stopping Power (MeV$\cdot$mg$^{-1}\cdot$cm$^{-2}$)')
		ax.set_xscale('log')
		ax.legend(loc=0)
		return _close_plot(f, ax, **kwargs)

	def plot(self, samples=None,  **kwargs):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if type(samples)==str:
			samples = [samples]
		f,ax = _init_plot(**kwargs)
		for sm in self.stack:
			if sm['name'] is not None:
				if samples is not None:
					if not any([re.match(s, sm['name']) for s in samples]):
						continue
				x, y = np.array([sm['bins'][:-1],sm['bins'][1:]]).T.flatten(), np.array([sm['flux'],sm['flux']]).T.flatten()
		
				ax.plot(x,y,label=sm['name'])

		ax.set_xlabel('Energy (MeV)')
		ax.set_ylabel('Flux (a.u.)')
		ax.legend(loc=0)
		return _close_plot(f, ax, **kwargs)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import numpy as np

from .dbmgr import get_cursor

class Element(object):
	"""Elemental data

	...
	
	Parameters
	----------
	x : type
		Description of parameter `x`.

	Attributes
	----------

	Methods
	-------

	Notes
	-----

	References
	----------

	Examples
	--------

	"""

	def __init__(self, element):
		self.element = element.title()
		self._meta = None

	@property
	def meta(self):
		if self._meta is None:
			from scipy.interpolate import interp1d

			self._meta = {}
			zg = get_cursor('ziegler')
			self._meta['Z'] = [str(i[2]).split(':')[0] for i in zg.execute('SELECT * FROM compounds WHERE compound=?',(self.element,))][0]
			self._meta['mass'], self._meta['density'] = [(i[1],i[2]) for i in zg.execute('SELECT * FROM weights WHERE Z=?',(self._meta['Z'],))][0]
			coeff = np.array([i[1:] for i in zg.execute('SELECT * FROM mass_coeff WHERE Z=? ORDER BY energy',(self._meta['Z'],))])
			self._meta['mass_coeff'] = interp1d(coeff[:,0], coeff[:,1], bounds_error=False, fill_value=0.0)
			self._meta['mass_coeff_en'] = interp1d(coeff[:,0], coeff[:,2], bounds_error=False, fill_value=0.0)
			db = get_cursor('decay')
			self._meta['abundances'] = {str(i[1]):i[12] for i in db.execute('SELECT * FROM chart WHERE element=? AND abundance>0.0',(self.element,))}
			self._meta['isotopes'] = sorted([i for i in self._meta['abundances']])

		return self._meta

	@property
	def Z(self):
		return self.meta['Z']

	@property
	def mass(self):
		return self.meta['mass']

	@property
	def isotopes(self):
		return self.meta['isotopes']

	@property
	def abundances(self):
		return self.meta['abundances']

	@property
	def mass_coeff(self):
		return self.meta['mass_coeff']

	@property
	def mass_coeff_en(self):
		return self.meta['mass_coeff_en']

	def attenuation(self, E, x=1.0):
		return np.exp(-self.mass_coeff(E)*self.density*x)

	def transmission(self, E, x=1.0):
		return 1.0-np.exp(-self.mass_coeff(E)*self.density*x)
	
	@property
	def density(self):
		return self.meta['density']
	
	
	


class Isotope(object):
	"""Isotopic data

	...
	
	Parameters
	----------
	x : type
		Description of parameter `x`.

	Attributes
	----------

	Methods
	-------

	Notes
	-----

	References
	----------

	Examples
	--------

	"""

	def __init__(self, istp):
		if istp=='1n' or istp=='1ng':
			self.element, self.A, self.isomer = 'n', 1, 'g'
		else:
			self.element = ''.join(re.findall('[A-Z]+',istp))
			if istp.startswith('nat'):
				self.A = 'nat'
			else:
				self.A = int(istp.split(self.element)[0])
			self.isomer = istp.split(self.element)[1]
		self.isotope = str(self.A)+self.element
		if self.isomer=='' and type(self.A)==int:
			self.isomer = 'g'
		if self.isomer=='m':
			self.isomer = 'm1'
		self.name = self.isotope+self.isomer
		self._meta = None
			
	@property
	def meta(self):
		if self._meta is None:
			self._meta = {}
			self.db = get_cursor('decay')
			for i in ['SFY','gm','el','bm','bp','al']:
				self._meta[i] = None
			Q = list(self.db.execute('SELECT * FROM chart WHERE isotope=?',(self.isotope,)))
			q = [i for i in Q if i[3]==self.isomer]
			if len(q):
				q = q[0]
			elif len(Q):
				q = Q[0]
			else:
				q = [0,0,0,0,'',0,None,None,None,None,None,None,None,None,None,None,'']
			self._meta['E_level'] = q[2]
			self._meta['J_pi'] = str(q[4]) if q[4] else '?'
			self._meta['Z'] = q[6]
			self._meta['N'] = q[7]
			self._meta['stable'] = bool(q[9])
			self._meta['t_half'] = q[10]
			self._meta['unc_t_half'] = q[11]
			self._meta['abundance'] = q[12]
			self._meta['unc_abundance'] = q[13]
			if self._meta['stable'] and self._meta['abundance'] is None:
				self._meta['abundance'] = 100.0
				self._meta['unc_abundance'] = 0.0
			self._meta['mass'] = q[14]
			self._meta['Delta'] = q[15]
			self._meta['decay_mode'] = list(map(lambda i:[float(p) if n==2 else p for n,p in enumerate(i.split(':'))], str(q[16]).split(',')))
			state = '' if len(Q)==1 else (self.isomer[0] if len(Q)==2 else self.isomer)
			self._meta['TeX'] = r'$^{'+str(self.A)+state+r'}$'+self.element.title()
		return self._meta
	
	@property
	def E_level(self):
		return self.meta['E_level']

	@property
	def J_pi(self):
		return self.meta['J_pi']

	@property
	def Z(self):
		return self.meta['Z']

	@property
	def N(self):
		return self.meta['N']

	@property
	def stable(self):
		return self.meta['stable']

	@property
	def mass(self):
		return self.meta['mass']

	@property
	def Delta(self):
		return self.meta['Delta']

	@property
	def TeX(self):
		return self.meta['TeX']

	def __str__(self):
		return self.name

	def half_life(self, units='s', unc=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.stable:
			return (np.inf,0.0) if unc else np.inf
		half_conv = {'ns':1e-9,'us':1e-6,'ms':1e-3,'s':1.0,'m':60.0,'h':3600.0,'d':86400.0,'y':31557.6E3,'ky':31557.6E6}[units]
		if unc:
			return self.meta['t_half']/half_conv,self.meta['unc_t_half']/half_conv
		return self.meta['t_half']/half_conv

	def decay_const(self, units='s', unc=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.stable:
			return (0.0,0.0) if unc else 0.0
		if unc:
			T2,uT2 = self.half_life(units, True)
			return np.log(2.0)/T2,np.log(2.0)*uT2/T2**2
		return np.log(2.0)/self.half_life(units)

	def optimum_units(self):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		opt = ['ns']
		for units in ['us','ms','s','m','h','d','y']:
			if self.half_life(units)>1.0:
				opt.append(units)
		return opt[-1]

	def abundance(self, unc=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if unc:
			return self.meta['abundance'], self.meta['unc_abundance']
		return self.meta['abundance']

	def get_SFY(self, unc=False, closest_SFY=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.meta['SFY'] is None:
			self.meta['SFY'] = [(str(i[1]),i[2],i[3]) for i in self.db.execute('SELECT * FROM SFY WHERE parent=?',(self.name,))]
		SFY = list(self.meta['SFY'])
		if len(SFY)==0 and closest_SFY:
			itps = list(set([str(i[0]) for i in self.db.execute('SELECT parent FROM SFY')]))
			dA = [abs(self.A-int(i[:3])) for i in itps]
			SFY = [(str(i[1]),i[2],i[3]) for i in self.db.execute('SELECT * FROM SFY WHERE parent=?',(itps[dA.index(min(dA))],))]
		if unc:
			return [[i[0], i[1], i[2]] for i in SFY]
		return [[i[0], i[1]] for i in SFY]

	def decay_products(self, closest_SFY=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		prods = []
		for (mode, product, br) in self.meta['decay_mode']:
			if product=='SFY':
				prods += [[i, br*y] for i,y in self.get_SFY(False, closest_SFY)]
			else:
				prods.append([product, br])
		return [[i, br] for i,br in prods if br>1E-8]

	def gammas(self, I_lim=[None,None], E_lim=[None,None], xrays=False, dE_511=3.5):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.meta['gm'] is None:
			self.meta['gm'] = [[float(i[3]),float(i[4]),float(i[5]),str(i[6])] for i in self.db.execute('SELECT * FROM gammas WHERE isotope=? AND isomer=?',(self.isotope,self.isomer))]
		gammas = list(self.meta['gm'])
		for n,L in enumerate([E_lim,I_lim]):
			if L[0] is not None:
				gammas = [g for g in gammas if g[n]>=L[0]]
			if L[1] is not None:
				gammas = [g for g in gammas if g[n]<=L[1]]
		if not xrays:
			gammas = [g for g in gammas if g[3]=='']
		return {l:[g[n] for g in gammas if abs(g[0]-511.0)>=dE_511] for n,l in enumerate(['E','I','dI','notes'])}

	def electrons(self,I_lim=(None,None),E_lim=(None,None),CE_only=False,Auger_only=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.meta['el'] is None:
			self.meta['el'] = [[float(i[3]),float(i[4]),float(i[5]),str(i[6])] for i in self.db.execute('SELECT * FROM electrons WHERE isotope=? AND isomer=?',(self.isotope,self.isomer))]
		electrons = list(self.meta['el'])
		for n,L in enumerate([E_lim,I_lim]):
			if L[0] is not None:
				electrons = [e for e in electrons if e[n]>=L[0]]
			if L[1] is not None:
				electrons = [e for e in electrons if e[n]<=L[1]]
		if CE_only:
			electrons = [e for e in electrons if e[3].startswith('CE')]
		if Auger_only:
			electrons = [e for e in electrons if e[3].startswith('Aug')]
		return {l:[e[n]for e in electrons] for n,l in enumerate(['E','I','dI','notes'])}

	def beta_minus(self,I_lim=(None,None),Endpoint_lim=(None,None)):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.meta['bm'] is None:
			self.meta['bm'] = [[float(i[3]),float(i[4]),float(i[5]),float(i[6])] for i in self.db.execute('SELECT * FROM beta_minus WHERE isotope=? AND isomer=?',(self.isotope,self.isomer))]
		betas = list(self.meta['bm'])
		for n,L in zip([3,1],[Endpoint_lim,I_lim]):
			if L[0] is not None:
				betas = [b for b in betas if b[n]>=L[0]]
			if L[1] is not None:
				betas = [b for b in betas if b[n]<=L[1]]
		return {l:[b[n] for b in betas] for n,l in enumerate(['muE','I','dI','endE'])}

	def beta_plus(self, I_lim=(None,None), Endpoint_lim=(None,None)):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.meta['bp'] is None:
			self.meta['bp'] = [[float(i[3]),float(i[4]),float(i[5]),float(i[6])] for i in self.db.execute('SELECT * FROM beta_plus WHERE isotope=? AND isomer=?',(self.isotope,self.isomer))]
		betas = list(self.meta['bp'])
		for n,L in zip([3,1],[Endpoint_lim,I_lim]):
			if L[0] is not None:
				betas = [b for b in betas if b[n]>=L[0]]
			if L[1] is not None:
				betas = [b for b in betas if b[n]<=L[1]]
		return {l:[b[n] for b in betas] for n,l in enumerate(['muE','I','dI','endE'])}

	def alphas(self, I_lim=(None,None), E_lim=(None,None)):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self.meta['al'] is None:
			self.meta['al'] = [[float(i[3]),float(i[4]),float(i[5])] for i in self.db.execute('SELECT * FROM alphas WHERE isotope=? AND isomer=?',(self.isotope,self.isomer))]
		alphas = list(self.meta['al'])
		for n,L in enumerate([E_lim,I_lim]):
			if L[0] is not None:
				alphas = [a for a in alphas if a[n]>=L[0]]
			if L[1] is not None:
				alphas = [a for a in alphas if a[n]<=L[1]]
		return {l:[a[n] for a in alphas] for n,l in enumerate(['E','I','dI'])}

	def dose_rate(self, activity=1.0, distance=30.0, units='R/hr'):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		def beta2(E_MeV, m_amu):
			return 1.0-(1.0/(1.0+(E_MeV/m_amu))**2)

		def e_range(E_keV):
			dEdx = lambda b2, t: 0.17*((np.log(3.61E5*t*np.sqrt(t+2)))+(0.5*(1-b2)*(1+t**2/8-(2*t+1)*np.log(2)))-4.312)/b2
			E = np.linspace(1E-12, E_keV*1E-3, 100)
			return np.trapz(1.0/((1.0+(E*7.22/800.0))*dEdx(beta2(E, 0.5109), E/0.5109)), E)

		def pos_range(E_keV):
			dEdx = lambda b2, t: 0.17*((np.log(3.61E5*t*np.sqrt(t+2)))+(np.log(2)-(b2/24)*(23+14/(t+2)+10/(t+2)**2+4/(t+2)**3))-4.312)/b2
			E = np.linspace(1E-12, E_keV*1E-3, 100)
			return np.trapz(1.0/((1.0+(E*7.22/800.0))*dEdx(beta2(E, 0.5109), E/0.5109)), E)

		def alpha_range(E_keV):
			dEdx = lambda b2: 0.17*4.0*((np.log(1.02E6*b2/(1.0-b2))-b2)-4.312)/b2
			return np.trapz(1.0/dEdx(beta2(np.linspace(1E-9, E_keV*1E-3, 100), 3.7284E3)), np.linspace(1E-9, E_keV*1E-3, 100))

		dose = {}
		gm = self.gammas(xrays=True, dE_511=0.0)
		al = self.alphas()
		bm = self.beta_minus()
		bp = self.beta_plus()
		el = self.electrons()

		dose['gammas'] = 1.4042E-12*np.sum(np.array(gm['E'])*np.array(gm['I']))*activity/distance**2
		dose['alphas'] = 5.2087E-14*np.sum([al['I'][n]*e/alpha_range(e) for n,e in enumerate(al['E'])])*activity/distance**2
		dose['beta_minus'] = 5.2087E-14*np.sum([bm['I'][n]*e/e_range(e) for n,e in enumerate(bm['muE'])])*activity/distance**2
		dose['beta_plus'] = 5.2087E-14*np.sum([bp['I'][n]*e/pos_range(e) for n,e in enumerate(bp['muE'])])*activity/distance**2
		dose['electrons'] = 5.2087E-14*np.sum([el['I'][n]*e/e_range(e) for n,e in enumerate(el['E'])])*activity/distance**2
		return dose





from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import datetime as dtm
from scipy.optimize import curve_fit

from .isotope import Isotope
from .spectroscopy import Spectrum
from .plotter import colors
from .plotter import _init_plot
from .plotter import _close_plot

#### FUTURE: Implement Irradiation history class

class DecayChain(object):
	"""Calculating activities within a decay chain using Bateman equations

	...

	Parameters
	----------
	x : type
		Description of parameter `x`.

	Attributes
	----------

	Methods
	-------

	Notes
	-----

	References
	----------

	Examples
	--------

	"""
	def __init__(self, parent, units='s', R=None, A0=None, time=None):
		self.units = units

		istps = [Isotope(parent)]
		self.isotopes = [istps[0].name]
		self.chain = [[istps[0].decay_const(units), [], []]]
		stable_chain = [False]

		while not all(stable_chain):
			for n in [n for n,ch in enumerate(stable_chain) if not ch]:
				stable_chain[n] = True
				for prod,br in istps[n].decay_products():
					I = Isotope(prod)
					if I.name in self.isotopes:
						self.chain[self.isotopes.index(I.name)][1].append(br)
						self.chain[self.isotopes.index(I.name)][2].append(n)
					else:
						istps.append(I)
						self.isotopes.append(I.name)
						self.chain.append([I.decay_const(units), [br], [n]])
						stable_chain.append(self.chain[-1][0]<1E-12)
		self.chain = np.array(self.chain, dtype=object)

		self.set_R(R)
		self.A0 = np.zeros(len(self.chain))
		self.update_A0(A0)

		self.time = time

		self._counts = [[] for i in self.chain]
		self._others = []
		self._prev = []
		self._R_fit = None
		self._A0_fit = None
		self._EoB = None

	def filter_name(self, istp):
		if istp[-1] in 'g0123456789':
			return istp
		return istp+('1' if istp.endswith('m') else 'g')

	def index(self, istp):
		return self.isotopes.index(self.filter_name(istp))

	def _get_branches(self, istp):
		if self.filter_name(istp) not in self.isotopes:
			return [], []
		m = self.index(istp)
		BR = [[0.0]]+[[r] for r in self.chain[m,1]]
		CH = [[m]]+[[n] for n in self.chain[m,2]]
		while not all([c[-1]==0 for c in CH]):
			BR = [BR[n]+[i] for n,c in enumerate(CH) for i in self.chain[c[-1],1]]
			CH = [c+[i] for c in CH for i in self.chain[c[-1],2]]
		BR = [np.array(r)[::-1] for n,r in enumerate(BR)]
		CH = [np.array(c)[::-1] for n,c in enumerate(CH)]
		return BR, CH

	def activity(self, istp, t=None, units=None, _R=None, _A0=None):
		"""Description

		...

		Parameters
		----------

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""
		
		t = t if t is not None else self.time
		if t is None:
			raise ValueError('Time must be specified.')
		t = np.asarray(t)
		A = np.zeros(len(t)) if t.shape else 0.0
		R_lm = 1.0
		if units is not None:
			conv = {'ns':1e-9,'us':1e-6,'ms':1e-3,'s':1.0,'m':60.0,'h':3600.0,'d':86400.0,'y':31557.6E3,'ky':31557.6E6}
			R_lm = conv[units]/conv[self.units]

		for m,(BR, chain) in enumerate(zip(*self._get_branches(istp))):
			lm = R_lm*self.chain[chain,0]
			A0 = self.A0[chain] if _A0 is None else _A0[chain]
			R = self.R[chain] if _R is None else _R[chain]
			n = len(chain)
			for i in range(n):
				if i==n-1 and m>0:
					continue

				A_i = lm[-1]*(A0[i]/lm[i])
				B_i = np.prod(lm[i:-1]*BR[i:-1])

				for j in range(i, n):
					K = np.arange(i, n)
					C_j = np.prod(lm[K[K!=j]]-lm[j])
					if np.any((lm[K[K!=j]]-lm[j])==0):
						C_j = np.where((lm[K[K!=j]]-lm[j])!=0,lm[K[K!=j]]-lm[j],1E-12)
						C_j = np.prod(C_j)
					A += A_i*B_i*np.exp(-lm[j]*t)/C_j
					if lm[j]>1E-12:
						A += R[i]*lm[-1]*B_i*(1.0-np.exp(-lm[j]*t))/(lm[j]*C_j)
					else:
						A += R[i]*lm[-1]*B_i*t/C_j

		for ot in self._others:
			if self.filter_name(istp) in ot.isotopes:
				A += ot.activity(istp, t, units=(self.units if units is None else units))
		return A


	def decays(self, istp, t_start, t_stop, units=None, _A0=None):
		"""Description

		...

		Parameters
		----------

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if np.any(self.R > 0):
			print('WARNING: decays during production not implemented.')
		t_start, t_stop = np.asarray(t_start), np.asarray(t_stop)
		D = np.zeros(len(t_start)) if t_start.shape else (np.zeros(len(t_stop)) if t_stop.shape else 0.0)
		R_lm = 1.0
		conv = {'ns':1e-9,'us':1e-6,'ms':1e-3,'s':1.0,'m':60.0,'h':3600.0,'d':86400.0,'y':31557600.0}
		if units is not None:
			R_lm = conv[units]/conv[self.units]

		for m,(BR, chain) in enumerate(zip(*self._get_branches(istp))):
			lm = R_lm*self.chain[chain,0]
			A0 = self.A0[chain] if _A0 is None else _A0[chain]
			n = len(chain)
			for i in range(n):
				if i==n-1 and m>0:
					continue

				A_i = lm[-1]*(A0[i]/lm[i])
				B_i = np.prod(lm[i:-1]*BR[i:-1])

				for j in range(i, len(chain)):
					K = np.arange(i, len(chain))
					C_j = np.prod(lm[K[K!=j]]-lm[j])
					if np.any((lm[K[K!=j]]-lm[j])==0):
						C_j = np.where((lm[K[K!=j]]-lm[j])!=0,lm[K[K!=j]]-lm[j],1E-12)
						C_j = np.prod(C_j)
					if lm[j]>1E-12:
						D += A_i*B_i*(np.exp(-lm[j]*t_start)-np.exp(-lm[j]*t_stop))/(lm[j]*C_j)
					else:
						D += A_i*B_i*(t_stop-t_start)/C_j
		D = D*conv[(self.units if units is None else units)]
		for ot in self._others:
			if self.filter_name(istp) in ot.isotopes:
				D += ot.decays(istp, t_start, t_stop, units=(self.units if units is None else units))
		return D

	def set_R(self, R):
		self.R = np.zeros(len(self.isotopes))
		if R is not None:
			if type(R)==dict:
				for istp in R:
					if self.filter_name(istp) in self.isotopes:
						self.R[self.index(istp)] += R[istp]
			else:
				self.R[0] += float(R)

	def update_A0(self, A0):
		if A0 is not None:
			if type(A0)==dict:
				for istp in A0:
					if self.filter_name(istp) in self.isotopes:
						self.A0[self.index(istp)] += A0[istp]
			else:
				self.A0[0] += float(A0)

	@property
	def EoB(self):
		return self._EoB

	@EoB.setter
	def EoB(self, eob):
		if type(eob)==str:
			eob = dtm.datetime.strptime(eob, '%m/%d/%Y %H:%M:%S')
		self._EoB = eob
	

	@property
	def counts(self):
		return self._counts

	@counts.setter
	def counts(self, N_c):
		if N_c is not None:
			if type(N_c)!=dict:
				N_c = {self.isotopes[0]:np.array(N_c)}
			for istp in N_c:
				if self.filter_name(istp) in self.isotopes:
					x = np.array(N_c[istp])
					i = self.index(istp)
					if x.shape:
						if len(x.shape)==1:
							x = np.array([x])
						if len(self._counts[i]):
							self._counts[i] = np.append(self._counts[i], x, axis=0)
						else:
							self._counts[i] = x

	@property
	def time(self):
		return self._time

	@time.setter
	def time(self, time):
		self._time = None
		if time is not None:
			t = np.asarray(time)
			if t.shape:
				self._time = t
			else:
				self._time = np.linspace(0.0, float(t), 1000)

	@property
	def A_meas(self):
		A_meas = []
		for n,ct in enumerate(self.counts):
			itp = self.isotopes[n]
			if len(ct):
				meas = np.array([self.calc_M(itp, itp, c[0], c[1])*c[2] for c in ct])
				if len(ct[0])==4:
					A_meas.append(np.column_stack((ct[:,0], meas, ct[:,3]*meas/ct[:,2])))
				else:
					A_meas.append(np.column_stack((ct[:,0], meas)))
			else:
				A_meas.append([])
		return A_meas
	
	def calc_L(self, i, t_m):
		"""Description

		...

		Parameters
		----------

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		a_0 = self.activity(i, 0.0)
		a_m = self.activity(i, t_m)
		if a_m>0:
			return a_0/a_m
		return 0.0

	def calc_M(self, i, j, t_start, t_stop):
		"""Description

		...

		Parameters
		----------

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		a_n = self.activity(i, t_start)
		d_m = self.decays(j, t_start, t_stop)
		if d_m>0:
			return a_n/d_m
		return 0.0

	def calc_P(self, i):
		"""Description

		...

		Parameters
		----------

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		a_0 = self.activity(i, 0.0)
		R_norm = self.R_norm[self.index(i)]
		if a_0>0:
			return R_norm/a_0
		return 0.0

	def calc_Q(self, i, t_m):
		"""Description

		...

		Parameters
		----------

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		a_m = self.activity(i, t_m)
		R_norm = self.R_norm[self.index(i)]
		if a_m>0:
			return R_norm/a_m
		return 0.0

	def fit_spectra(self, spectra, db=None, max_unc=0.15, EoB=None):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if EoB is not None:
			self.EoB = EoB
		if self.EoB is None:
			print('WARNING: EoB must be set.')
			return
		counts = {}
		conv = {'ns':1e-9,'us':1e-6,'ms':1e-3,'s':1.0,'m':60.0,'h':3600.0,'d':86400.0,'y':31557600.0}
		for sp in spectra:
			if type(sp)==str:
				sp = Spectrum(sp, db)
			start_time = (sp.meta['start_time']-self.EoB).total_seconds()/conv[self.units]
			end_time = start_time+(sp.meta['real_time']/conv[self.units])
			if not sp.meta['istp']:
				sp.meta['istp'] = self.isotopes
			for n,p in sp.peaks.iterrows():
				if not (p['converged'] and p['unc_decays']<max_unc*p['decays']):
					continue
				if self.filter_name(p['isotope']) not in self.isotopes:
					continue
				if p['isotope'] in counts:
					counts[p['isotope']].append([start_time, end_time, p['decays'], p['unc_decays']])
				else:
					counts[p['isotope']] = [[start_time, end_time, p['decays'], p['unc_decays']]]
		self.counts = counts

	@property
	def R_norm(self):
		R_p,t = [p.R for p in self._prev if p.R.any()],[p.time[-1] for p in self._prev if p.R.any()]
		return np.dot(np.array(R_p).T,np.array(t))/np.sum(t)

	def fit_R(self, istp=None, _update=True, unc=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self._R_fit is not None:
			if unc:
				if istp is None:
					return self._R_fit, self._unc_R_fit
				return self._R_fit[self.index(istp)], self._unc_R_fit[self.index(istp)]

			if istp is None:
				return self._R_fit
			return self._R_fit[self.index(istp)]

		R_norm = self.R_norm
		nz = np.where(R_norm>0)[0]
		Y = []
		dY = []
		X = []
		time = []
		itp = []
		for n,ct in enumerate(self.counts):
			if len(ct):
				Y += ct[:,2].tolist()
				if len(ct[0])==4:
					dY += ct[:,3].tolist()
				X += [np.zeros(len(nz)) for i in ct]
				time += ct[:,:2].tolist()
				itp += (n*np.ones(len(ct),dtype=int)).tolist()

		Y = np.array(Y)
		dY = np.array(dY) if len(dY) else np.ones(len(Y))
		X = np.asarray(X)

		I = np.arange(len(self.isotopes))
		for m,z in enumerate(nz):
			A0 = np.copy(self._prev[0].A0) if len(self._prev) else np.zeros(len(self.isotopes))
			for p in self._prev:
				R_p = p.R/np.where(R_norm>0, R_norm, 1.0)
				R_p[I!=z] = 0.0
				A_old = np.copy(A0)
				for n in I:
					A0[n] = self.activity(self.isotopes[n], p.time[-1], _R=R_p, _A0=A_old)
			X[:,m] = np.array([self.decays(self.isotopes[itp[n]], time[n][0], time[n][1], _A0=A0) for n in range(len(X))])
		
		func = lambda X_f, *R_f: np.dot(X_f, np.asarray(R_f).T)
		fit, cov = curve_fit(func, X, Y, sigma=dY, p0=R_norm[nz], bounds=([0.0 for i in nz],[np.inf for i in nz]))
		self._R_fit = np.zeros(len(R_norm))
		self._R_fit[nz] = fit
		self._unc_R_fit = np.zeros(len(R_norm))
		self._unc_R_fit[nz] = (np.sqrt(np.diag(cov)) if np.all(np.isfinite(cov)) else np.diag(cov))

		if _update:
			A0 = np.copy(self._prev[0].A0) if len(self._prev) else np.zeros(len(self.isotopes))
			for p in self._prev:
				p.A0 = np.copy(A0)
				p.R = p.R*self._R_fit/np.where(R_norm>0, R_norm, 1.0)
				for n in I:
					A0[n] = p.activity(self.isotopes[n], p.time[-1])
			self.A0 = np.copy(A0)

		if unc:
			if istp is None:
				return self._R_fit, self._unc_R_fit
			return self._R_fit[self.index(istp)], self._unc_R_fit[self.index(istp)]

		if istp is None:
			return self._R_fit
		return self._R_fit[self.index(istp)]

	def fit_A0(self, istp=None, _update=True, unc=False):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if self._A0_fit is not None:
			if unc:
				if istp is None:
					return self._A0_fit, self._unc_A0_fit
				return self._A0_fit[self.index(istp)], self._unc_A0_fit[self.index(istp)]

			if istp is None:
				return self._A0_fit
			return self._A0_fit[self.index(istp)]

		if len(self._prev):
			print('WARNING: Previous activities and production rates will not be fit.')

		nz = np.where(self.A0>0)[0]

		Y = []
		dY = []
		X = []
		time = []
		itp = []
		for n,ct in enumerate(self.counts):
			if len(ct):
				Y += ct[:,2].tolist()
				if len(ct[0])==4:
					dY += ct[:,3].tolist()
				X += [np.zeros(len(nz)) for i in ct]
				time += ct[:,:2].tolist()
				itp += (n*np.ones(len(ct),dtype=int)).tolist()

		Y = np.array(Y)
		dY = np.array(dY) if len(dY) else np.ones(len(Y))
		X = np.asarray(X)

		I = np.arange(len(self.isotopes))
		for z in nz:
			A0 = np.where(I!=z, np.zeros(len(self.isotopes)), 1.0)
			X[:,z] = np.array([self.decays(self.isotopes[itp[n]], time[n][0], time[n][1], _A0=A0) for n in range(len(X))])

		func = lambda X_f, *R_f: np.dot(X_f, np.asarray(R_f).T)
		fit, cov = curve_fit(func, X, Y, sigma=dY, p0=self.A0[nz], bounds=([0.0 for i in nz],[np.inf for i in nz]))
		self._A0_fit = np.copy(self.A0)
		self._A0_fit[self.A0>0] = fit
		self._unc_A0_fit = np.zeros(len(self.A0))
		self._unc_A0_fit[self.A0>0] = (np.sqrt(np.diag(cov)) if np.all(np.isfinite(cov)) else np.diag(cov))

		if _update:
			self.A0 = np.copy(self._A0_fit)

		if unc:
			if istp is None:
				return self._A0_fit, self._unc_A0_fit
			return self._A0_fit[self.index(istp)], self._unc_A0_fit[self.index(istp)]

		if istp is None:
			return self._A0_fit
		return self._A0_fit[self.index(istp)]


	def __add__(self, other):
		self._others.append(other)
		return self

	def append(self, other, time=None):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		if time is not None:
			self.time = time
		if self.time is None:
			raise ValueError('Decay Chain time is not set.')
		if self.isotopes[0]!=other.isotopes[0]:
			raise ValueError('Decay chains must have same parent.')
		if self.units!=other.units:
			raise ValueError('Units must be the same.')

		tmp = [self.R, self.A0, self.time, self._others, self.counts]
		new_A0 = {i:self.activity(i, self.time[-1]) for n,i in enumerate(self.isotopes) if self.chain[n,0]>1E-12}
		self.R = other.R
		self.A0 = other.A0
		self.update_A0(new_A0)
		self.time = other.time
		self._others = other._others
		self._counts = other.counts
		other.R = tmp[0]
		other.A0 = tmp[1]
		other.time = tmp[2]
		other._others = tmp[3]
		other._counts = tmp[4]
		self._prev += other._prev
		self._prev.append(other)
		other._prev = []

	def plot(self, time=None, N_plot=None, N_label=10, **kwargs):
		"""Description

		...

		Parameters
		----------
		x : type
			Description of parameter `x`.

		Returns
		-------

		Notes
		-----

		References
		----------

		Examples
		--------

		"""

		f, ax = _init_plot(**kwargs)
		if time is not None:
			self.time = time
		if self.time is not None:
			t = self.time
		else:
			t = np.linspace(0, 5.0*np.log(2)/self.chain[0,0], 1000)
		if N_plot is None:
			N_plot = len(self.isotopes)
		A_meas = self.A_meas
		ordr = int(np.floor(np.log10(np.average(self.activity(self.isotopes[0], t)))/3.0))
		lb_or = {-4:'p',-3:'n',-2:r'$\mu$',-1:'m',0:'',1:'k',2:'M',3:'G',4:'T'}[ordr]
		mult = 10.0**(-3*ordr)
		for n,istp in enumerate(self.isotopes):
			if self.chain[n,0]>1E-12 and n<N_plot:
				dt = 0.0
				tm, A = np.array([]),np.array([])
				for p in self._prev:
					dt += p.time[-1]
					t_p, A_p = p.time, p.activity(istp)
					tm, A = np.append(tm, t_p+(tm[-1] if len(tm) else 0.0)), np.append(A, A_p)
				tm, A = np.append(tm-dt, t), np.append(A, self.activity(istp, t))
				label = Isotope(istp).TeX if n<N_label else None
				line, = ax.plot(tm, mult*A, label=label)
				if len(A_meas[n]):
					am = A_meas[n]
					if len(am[0])==3:
						ax.errorbar(am[:,0], mult*am[:,1], yerr=mult*am[:,2], ls='None', marker='o', color=line.get_color())
					else:
						ax.plot(am[:,0], mult*am[:,1], ls='None', marker='o', color=line.get_color())

		ax.set_xlabel('Time ({})'.format(self.units))
		ax.set_ylabel('Activity ({}Bq)'.format(lb_or))
		ax.legend(loc=0)
		return _close_plot(f, ax, default_log=False, **kwargs)




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sqlite3
import npat

path = lambda db: os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)),'data',db))

DECAY_connection = None
ZIEGLER_connection = None
ENDF_connection = None
TENDL_connection = None
TENDL_rpn_connection = None
TENDL_rpp_connection = None
TENDL_rpd_connection = None
IRDFF_connection = None
CPR_connection = None

def download(db='decay',f=False):
	"""Some nice color maps.

	...

	Parameters
	----------

	Returns
	-------

	Notes
	-----

	References
	----------

	Examples
	--------

	"""

	db = db.lower()
	if db in ['all','*']:
		d = ['decay','ziegler','endf','tendl','tendl_n_rp','tendl_p_rp','tendl_d_rp','IRDFF','iaea_monitors']
	elif db in ['decay']:
		d = ['decay']
	elif db in ['ziegler']:
		d = ['ziegler']
	elif db in ['endf']:
		d = ['endf']
	elif db in ['tendl_n_rp','tendl_nrp','tendl_n','nrp','rpn']:
		d = ['tendl_n_rp']
	elif db in ['tendl_p_rp','tendl_prp','tendl_p','prp','rpp']:
		d = ['tendl_p_rp']
	elif db in ['tendl_d_rp','tendl_drp','tendl_d','drp','rpd']:
		d = ['tendl_d_rp']
	elif db in ['irdff']:
		d = ['IRDFF']
	elif db in ['iaea','iaea-cpr','iaea-monitor','cpr','iaea_cpr','iaea_monitor','medical','iaea-medical','iaea_medical']:
		d = ['iaea_monitors']
	else:
		print('db={} not recognized.'.format(db))
		return

	addr = {'decay':'wwd6b1gk2ge5tgt', 'endf':'tkndjqs036piojm', 'tendl':'zkoi6t2jicc9yqs', 'tendl_d_rp':'x2vfjr7uv7ffex5', 'tendl_n_rp':'n0jjc0dv61j9of9',
				'tendl_p_rp':'ib2a5lrhiwkcro5', 'ziegler':'kq07684wtp890v5','iaea_monitors':'lzn8zs6y8zu3v0s','IRDFF':'34sgcvt8n57b0aw'}
	
	if not os.path.isdir(path('')):
		os.mkdir(path(''))

	try:
		import urllib2
	except:
		import urllib.request as urllib2

	for i in d:
		fnm = i+'.db'
		if (not os.path.isfile(path(fnm))) or f:
			
			try:
				print('Downloading {}'.format(fnm))
				with open(path(fnm),'wb') as f:
					f.write(urllib2.urlopen('https://www.dropbox.com/s/{0}/{1}?dl=1'.format(addr[i],fnm)).read())
			except Exception as e:
					print(e)
		else:
			print("{0}.db already installed. Run npat.download(db='{0}',f=True) to override.".format(i))



def get_connection(db='decay'):
	"""Some nice color maps.

	...

	Parameters
	----------

	Returns
	-------

	Notes
	-----

	References
	----------

	Examples
	--------

	"""

	db = db.lower().replace('.db','')

	if db in ['decay']:
		global DECAY_connection
		if DECAY_connection is None:
			DECAY_connection = sqlite3.connect(path('decay.db'))

		return DECAY_connection

	elif db in ['ziegler']:
		global ZIEGLER_connection
		if ZIEGLER_connection is None:
			ZIEGLER_connection = sqlite3.connect(path('ziegler.db'))

		return ZIEGLER_connection

	elif db in ['endf']:
		global ENDF_connection
		if ENDF_connection is None:
			ENDF_connection = sqlite3.connect(path('endf.db'))

		return ENDF_connection

	elif db in ['tendl']:
		global TENDL_connection
		if TENDL_connection is None:
			TENDL_connection = sqlite3.connect(path('tendl.db'))

		return TENDL_connection

	elif db in ['tendl_n_rp','tendl_nrp','tendl_n','nrp','rpn']:
		global TENDL_rpn_connection
		if TENDL_rpn_connection is None:
			TENDL_rpn_connection = sqlite3.connect(path('tendl_n_rp.db'))

		return TENDL_rpn_connection

	elif db in ['tendl_p_rp','tendl_prp','tendl_p','prp','rpp']:
		global TENDL_rpp_connection
		if TENDL_rpp_connection is None:
			TENDL_rpp_connection = sqlite3.connect(path('tendl_p_rp.db'))

		return TENDL_rpp_connection

	elif db in ['tendl_d_rp','tendl_drp','tendl_d','drp','rpd']:
		global TENDL_rpd_connection
		if TENDL_rpd_connection is None:
			TENDL_rpd_connection = sqlite3.connect(path('tendl_d_rp.db'))

		return TENDL_rpd_connection	

	elif db in ['irdff']:
		global IRDFF_connection
		if IRDFF_connection is None:
			IRDFF_connection = sqlite3.connect(path('IRDFF.db'))

		return IRDFF_connection

	elif db in ['iaea','iaea-cpr','iaea-monitor','cpr','iaea_cpr','iaea_monitor','medical','iaea-medical','iaea_medical']:
		global CPR_connection
		if CPR_connection is None:
			CPR_connection = sqlite3.connect(path('iaea_monitors.db'))

		return CPR_connection	


def get_cursor(db='decay'):
	"""Some nice color maps.

	...

	Parameters
	----------

	Returns
	-------

	Notes
	-----

	References
	----------

	Examples
	--------

	"""
	
	conn = get_connection(db)
	return conn.cursor()
